# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Helper functions to prepare ASTRA algorithms.

This module contains utility functions to convert data structures from the
ODL geometry representation to ASTRA's data structures, including:

* volume geometries
* projection geometries
* create vectors from geometries
* data arrays
* projectors
* algorithm configuration dictionaries

`ASTRA documentation on Sourceforge
<https://sourceforge.net/p/astra-toolbox/wiki>`_.

`ASTRA on GitHub
<https://github.com/astra-toolbox/>`_.
"""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()

try:
    import astra
    ASTRA_AVAILABLE = True
except ImportError:
    ASTRA_AVAILABLE = False
import numpy as np

from odl.discr import DiscreteLp, DiscreteLpElement
from odl.tomo.geometry import (
    Geometry, Parallel2dGeometry, DivergentBeamGeometry, ParallelGeometry,
    FlatDetector)


__all__ = ('ASTRA_AVAILABLE', 'astra_volume_geometry',
           'astra_projection_geometry', 'astra_data', 'astra_projector',
           'astra_algorithm',
           'astra_conebeam_3d_geom_to_vec',
           'astra_conebeam_2d_geom_to_vec',
           'astra_parallel_3d_geom_to_vec')


def astra_volume_geometry(discr_reco):
    """Create an ASTRA volume geometry from the discretized domain.

    From the ASTRA documentation:

    In all 3D geometries, the coordinate system is defined around the
    reconstruction volume. The center of the reconstruction volume is the
    origin, and the sides of the voxels in the volume have length 1.

    All dimensions in the projection geometries are relative to this unit
    length.


    Parameters
    ----------
    discr_reco : `DiscreteLp`
        Discretization of an L2 space on the reconstruction domain.
        It must be 2- or 3-dimensional and uniformly discretized.

    Returns
    -------
    astra_geom : dict
        The ASTRA volume geometry

    Raises
    ------
    NotImplementedError
        if the cell sizes are not the same in each dimension
    """
    # TODO: allow other discretizations?
    if not isinstance(discr_reco, DiscreteLp):
        raise TypeError('`discr_reco` {!r} is not a DiscreteLp instance'
                        ''.format(discr_reco))

    if not discr_reco.is_uniform:
        raise ValueError('`discr_reco` {} is not uniformly discretized')

    vol_shp = discr_reco.partition.shape
    vol_min = discr_reco.partition.min_pt
    vol_max = discr_reco.partition.max_pt

    if discr_reco.ndim == 2:
        # ASTRA does in principle support custom minimum and maximum
        # values for the volume extent, but projector creation fails
        # if voxels are non-isotropic. We raise an exception here in
        # the meanwhile.
        if not np.allclose(discr_reco.partition.cell_sides[1:],
                           discr_reco.partition.cell_sides[:-1]):
            raise NotImplementedError('non-isotropic voxels not supported by '
                                      'ASTRA')
        # given a 2D array of shape (x, y), a volume geometry is created as:
        #    astra.create_vol_geom(x, y, y_min, y_max, x_min, x_max)
        # yielding a dictionary:
        #   'GridColCount': y,
        #   'GridRowCount': x
        #   'WindowMaxX': y_max
        #   'WindowMaxY': x_max
        #   'WindowMinX': y_min
        #   'WindowMinY': x_min
        vol_geom = astra.create_vol_geom(vol_shp[0], vol_shp[1],
                                         vol_min[1], vol_max[1],
                                         vol_min[0], vol_max[0])
    elif discr_reco.ndim == 3:
        # Non-isotropic voxels are not yet supported in 3d ASTRA
        if not np.allclose(discr_reco.partition.cell_sides[1:],
                           discr_reco.partition.cell_sides[:-1]):
            # TODO: for parallel geometries, one can work around this issue
            raise NotImplementedError('non-isotropic voxels not supported by '
                                      'ASTRA')
        # given a 3D array of shape (x, y, z), a volume geometry is created as:
        #    astra.create_vol_geom(y, z, x, )
        # yielding a dictionary:
        #   'GridColCount': z
        #   'GridRowCount': y
        #   'GridSliceCount': x
        #   'WindowMinX': z_max
        #   'WindowMaxX': z_max
        #   'WindowMinY': y_min
        #   'WindowMaxY': y_min
        #   'WindowMinZ': x_min
        #   'WindowMaxZ': x_min
        vol_geom = astra.create_vol_geom(vol_shp[1], vol_shp[2], vol_shp[0],
                                         vol_min[2], vol_max[2],
                                         vol_min[1], vol_max[1],
                                         vol_min[0], vol_max[0])
    else:
        raise ValueError('{}-dimensional volume geometries not supported '
                         'by ASTRA'.format(discr_reco.ndim))
    return vol_geom


def astra_conebeam_3d_geom_to_vec(geometry):
    """Create vectors for ASTRA projection geometries from ODL geometry.

    The 3D vectors are used to create an ASTRA projection geometry for
    cone beam geometries ('cone_vec') with helical acquisition curves.

    Output vectors:

    Each row of vectors corresponds to a single projection, and consists of:
        ( srcX, srcY, srcZ, dX, dY, dZ, uX, uY, uZ, vX, vY, vZ ):
        src : the ray source
        d   : the center of the detector
        u   : the vector from detector pixel (0,0) to (0,1)
        v   : the vector from detector pixel (0,0) to (1,0)

    Parameters
    ----------
    geometry : `Geometry`
        The ODL geometry instance used to create the ASTRA geometry

    Returns
    -------
    vectors : `numpy.ndarray`
        Numpy array of shape ``(number of angles, 12)``
    """

    angles = geometry.angles
    vectors = np.zeros((angles.size, 12))

    for ang_idx, angle in enumerate(angles):
        rot_matrix = geometry.rotation_matrix(angle)

        # source position
        vectors[ang_idx, 0:3] = geometry.src_position(angle)

        # center of detector
        mid_pt = geometry.det_params.mid_pt
        vectors[ang_idx, 3:6] = geometry.det_point_position(angle, mid_pt)

        # vector from detector pixel (0,0) to (0,1)
        unit_vecs = geometry.detector.axes
        strides = geometry.det_grid.stride
        vectors[ang_idx, 6:9] = rot_matrix.dot(unit_vecs[0] * strides[0])
        vectors[ang_idx, 9:12] = rot_matrix.dot(unit_vecs[1] * strides[1])

    # Astra order, needed for data to match what we expect from astra.
    # Astra has a different axis convention to ODL (z, y, x), so we need
    # to adapt to this by changing the order
    newind = []
    for i in range(4):
        newind += [2 + 3 * i, 1 + 3 * i, 0 + 3 * i]
    vectors = vectors[:, newind]

    return vectors


def astra_conebeam_2d_geom_to_vec(geometry):
    """Create vectors for ASTRA projection geometries from ODL geometry.

    The 2D vectors are used to create an ASTRA projection geometry for
    cone beam geometries ('flat_vec') with helical acquisition curves.

    Output vectors:

    Each row of vectors corresponds to a single projection, and consists of:
        ( srcX, srcY, dX, dY, uX, uY )
        src : the ray source
        d : the center of the detector
        u : the vector between the centers of detector pixels 0 and 1

    Parameters
    ----------
    geometry : `Geometry`
        The ODL geometry instance used to create the ASTRA geometry

    Returns
    -------
    vectors : `numpy.ndarray`
        Numpy array of shape ``(number of angles, 6)``
    """

    angles = geometry.angles
    vectors = np.zeros((angles.size, 6))

    for ang_idx, angle in enumerate(angles):
        rot_matrix = geometry.rotation_matrix(angle)

        # source position
        vectors[ang_idx, 0:2] = geometry.src_position(angle)

        # center of detector
        mid_pt = geometry.det_params.mid_pt
        vectors[ang_idx, 2:4] = geometry.det_point_position(angle, mid_pt)

        # vector from detector pixel (0) to (1)
        unit_vec = geometry.detector.axis
        strides = geometry.det_grid.stride
        vectors[ang_idx, 4:6] = rot_matrix.dot(unit_vec * strides[0])

    # Astra order, needed for data to match what we expect from astra.
    # Astra has a different axis convention to ODL (z, y, x), so we need
    # to adapt to this by changing the order
    newind = []
    for i in range(3):
        newind += [1 + 2 * i, 0 + 2 * i]
    vectors = vectors[:, newind]

    return vectors


def astra_parallel_3d_geom_to_vec(geometry):
    """Create vectors for ASTRA projection geometries from ODL geometry.

    The 3D vectors are used to create an ASTRA projection geometry for
    parallel beam geometries ('parallel3d_vec').

    Output vectors:

    Each row of vectors corresponds to a single projection, and consists of:
        ( rayX, rayY, rayZ, dX, dY, dZ, uX, uY, uZ, vX, vY, vZ )
        ray : the ray direction
        d   : the center of the detector
        u   : the vector from detector pixel (0,0) to (0,1)
        v   : the vector from detector pixel (0,0) to (1,0)

    Parameters
    ----------
    geometry : `Geometry`
        The ODL geometry instance used to create the ASTRA geometry

    Returns
    -------
    vectors : `numpy.ndarray`
        Numpy array of shape ``(number of angles, 12)``
    """

    angles = geometry.angles
    vectors = np.zeros((angles.size, 12))

    for ang_idx, angle in enumerate(angles):
        rot_matrix = geometry.rotation_matrix(angle)

        mid_pt = geometry.det_params.mid_pt

        # source position
        vectors[ang_idx, 0:3] = geometry.det_to_src(angle, mid_pt)

        # center of detector
        vectors[ang_idx, 3:6] = geometry.det_point_position(angle, mid_pt)

        # vector from detector pixel (0,0) to (0,1)
        unit_vecs = geometry.detector.axes
        strides = geometry.det_grid.stride
        vectors[ang_idx, 6:9] = rot_matrix.dot(unit_vecs[0] * strides[0])
        vectors[ang_idx, 9:12] = rot_matrix.dot(unit_vecs[1] * strides[1])

    # Astra order, needed for data to match what we expect from astra.
    # Astra has a different axis convention to ODL (z, y, x), so we need
    # to adapt to this by changing the order
    new_ind = []
    for i in range(4):
        new_ind += [2 + 3 * i, 1 + 3 * i, 0 + 3 * i]
    vectors = vectors[:, new_ind]

    return vectors


def astra_projection_geometry(geometry):
    """Create an ASTRA projection geometry from an ODL geometry object.

    As of ASTRA version 1.7, the length values are not required any more to be
    rescaled for 3D geometries and non-unit (but isotropic) voxel sizes.

    Parameters
    ----------
    geometry : `Geometry`
        Object from which to create the ASTRA projection geometry.

    Returns
    -------
    proj_geom : dict
        Dictionary defining the ASTRA projection geometry.
    """
    if not isinstance(geometry, Geometry):
        raise TypeError('`geometry` {!r} is not a `Geometry` instance'
                        ''.format(geometry))

    if 'astra' in geometry.implementation_cache:
        # Shortcut, reuse already computed value.
        return geometry.implementation_cache['astra']

    if not geometry.det_partition.is_uniform:
        raise ValueError('non-uniform detector sampling is not supported')

    # As of ASTRA version 1.7beta the volume width can be specified in the
    # volume geometry creator also for 3D geometries. For version < 1.7
    # this was possible only for 2D geometries. Thus, standard ASTRA
    # projection geometries do not have to be rescaled any more by the (
    # isotropic) voxel size.
    if isinstance(geometry, Parallel2dGeometry):
        # TODO: change to parallel_vec when available
        det_width = geometry.det_partition.cell_sides[0]
        det_count = geometry.detector.size
        angles = geometry.angles
        proj_geom = astra.create_proj_geom('parallel', det_width, det_count,
                                           angles)

    elif (isinstance(geometry, DivergentBeamGeometry) and
          isinstance(geometry.detector, FlatDetector) and
          geometry.ndim == 2):
        det_count = geometry.detector.size
        vec = astra_conebeam_2d_geom_to_vec(geometry)
        proj_geom = astra.create_proj_geom('fanflat_vec', det_count, vec)

    elif (isinstance(geometry, ParallelGeometry) and
          isinstance(geometry.detector, FlatDetector) and
          geometry.ndim == 3):
        det_row_count = geometry.det_partition.shape[1]
        det_col_count = geometry.det_partition.shape[0]
        vec = astra_parallel_3d_geom_to_vec(geometry)
        proj_geom = astra.create_proj_geom('parallel3d_vec', det_row_count,
                                           det_col_count, vec)

    elif (isinstance(geometry, DivergentBeamGeometry) and
          isinstance(geometry.detector, FlatDetector) and
          geometry.ndim == 3):
        det_row_count = geometry.det_partition.shape[1]
        det_col_count = geometry.det_partition.shape[0]
        vec = astra_conebeam_3d_geom_to_vec(geometry)
        proj_geom = astra.create_proj_geom('cone_vec', det_row_count,
                                           det_col_count, vec)
    else:
        raise NotImplementedError('unknown ASTRA geometry type {!r}'
                                  ''.format(geometry))

    if 'astra' not in geometry.implementation_cache:
        # Save computed value for later
        geometry.implementation_cache['astra'] = proj_geom

    return proj_geom


def astra_data(astra_geom, datatype, data=None, ndim=2, allow_copy=False):
    """Create an ASTRA data structure.

    Parameters
    ----------
    astra_geom : dict
        ASTRA geometry object for the data creator, must correspond to the
        given data type
    datatype : {'volume', 'projection'}
        Type of the data container
    data : `DiscreteLpElement`, optional
        Data for the initialization of the data structure. If ``None``,
        an ASTRA data object filled with zeros is created.
    ndim : {2, 3}, optional
        Dimension of the data. If ``data`` is not ``None``, this parameter
        has no effect.
    allow_copy : `bool`, optional
        True if copying ``data`` should be allowed. This means that anything
        written by ASTRA to the returned structure will not be written to
        ``data``.

    Returns
    -------
    id : int
        ASTRA internal ID for the new data structure
    """
    if data is not None:
        if isinstance(data, DiscreteLpElement):
            ndim = data.space.ndim
        elif isinstance(data, np.ndarray):
            ndim = data.ndim
        else:
            raise TypeError('`data` {!r} is neither DiscreteLpElement '
                            'instance nor a `numpy.ndarray`'.format(data))
    else:
        ndim = int(ndim)

    if datatype == 'volume':
        astra_dtype_str = '-vol'
    elif datatype == 'projection':
        astra_dtype_str = '-sino'
    else:
        raise ValueError('`datatype` {!r} not understood'.format(datatype))

    # Get the functions from the correct module
    if ndim == 2:
        link = astra.data2d.link
        create = astra.data2d.create
    elif ndim == 3:
        link = astra.data3d.link
        create = astra.data3d.create
    else:
        raise ValueError('{}-dimensional data structures not supported'
                         ''.format(ndim))

    # ASTRA checks if data is c-contiguous and aligned
    if data is not None:
        if allow_copy:
            data_array = np.asarray(data, dtype='float32', order='C')
            return link(astra_dtype_str, astra_geom, data_array)
        else:
            if isinstance(data, np.ndarray):
                return link(astra_dtype_str, astra_geom, data)
            elif data.ntuple.impl == 'numpy':
                return link(astra_dtype_str, astra_geom, data.asarray())
            else:
                # Something else than NumPy data representation
                raise NotImplementedError('ASTRA supports data wrapping only '
                                          'for `numpy.ndarray` instances, got '
                                          '{!r}'.format(data))
    else:
        return create(astra_dtype_str, astra_geom)


def astra_projector(vol_interp, astra_vol_geom, astra_proj_geom, ndim, impl):
    """Create an ASTRA projector configuration dictionary.

    Parameters
    ----------
    vol_interp : {'nearest', 'linear'}
        Interpolation type of the volume discretization
    astra_vol_geom : dict
        ASTRA volume geometry dictionary
    astra_proj_geom : dict
        ASTRA projection geometry dictionary
    ndim : {2, 3}
        Number of dimensions of the projector
    impl : {'cpu', 'cuda'}
        Implementation of the projector

    Returns
    -------
    proj_id : int
        ASTRA reference ID to the ASTRA dict with initialized 'type' key
    """
    if vol_interp not in ('nearest', 'linear'):
        raise ValueError("`vol_interp` '{}' not understood"
                         ''.format(vol_interp))
    impl = str(impl).lower()
    if impl not in ('cpu', 'cuda'):
        raise ValueError("`impl` '{}' not understood"
                         ''.format(impl))

    if 'type' not in astra_proj_geom:
        raise ValueError('invalid projection geometry dict {}'
                         ''.format(astra_proj_geom))
    if ndim == 3 and impl == 'cpu':
        raise ValueError('3D projectors not supported on CPU')

    ndim = int(ndim)

    proj_type = astra_proj_geom['type']
    if proj_type not in ('parallel', 'fanflat', 'fanflat_vec',
                         'parallel3d', 'parallel3d_vec', 'cone', 'cone_vec'):
        raise ValueError('invalid geometry type {!r}'.format(proj_type))

    # Mapping from interpolation type and geometry to ASTRA projector type.
    # "I" means probably mathematically inconsistent. Some projectors are
    # not implemented, e.g. CPU 3d projectors in general.
    type_map_cpu = {'parallel': {'nearest': 'line',
                                 'linear': 'linear'},  # I
                    'fanflat': {'nearest': 'line_fanflat',
                                'linear': 'line_fanflat'},  # I
                    'parallel3d': {'nearest': 'linear3d',  # I
                                   'linear': 'linear3d'},  # I
                    'cone': {'nearest': 'linearcone',  # I
                             'linear': 'linearcone'}}  # I
    type_map_cpu['fanflat_vec'] = type_map_cpu['fanflat']
    type_map_cpu['parallel3d_vec'] = type_map_cpu['parallel3d']
    type_map_cpu['cone_vec'] = type_map_cpu['cone']

    # GPU algorithms not necessarily require a projector, but will in future
    # releases making the interface more coherent regarding CPU and GPU
    type_map_cuda = {'parallel': 'cuda',  # I
                     'parallel3d': 'cuda3d'}  # I
    type_map_cuda['fanflat'] = type_map_cuda['parallel']
    type_map_cuda['fanflat_vec'] = type_map_cuda['fanflat']
    type_map_cuda['cone'] = type_map_cuda['parallel3d']
    type_map_cuda['parallel3d_vec'] = type_map_cuda['parallel3d']
    type_map_cuda['cone_vec'] = type_map_cuda['cone']

    # create config dict
    proj_cfg = {}
    if impl == 'cpu':
        proj_cfg['type'] = type_map_cpu[proj_type][vol_interp]
    else:  # impl == 'cuda'
        proj_cfg['type'] = type_map_cuda[proj_type]
    proj_cfg['VolumeGeometry'] = astra_vol_geom
    proj_cfg['ProjectionGeometry'] = astra_proj_geom

    if ndim == 2:
        return astra.projector.create(proj_cfg)
    else:
        return astra.projector3d.create(proj_cfg)


def astra_algorithm(direction, ndim, vol_id, sino_id, proj_id, impl):
    """Create an ASTRA algorithm object to run the projector.

    Parameters
    ----------
    direction : {'forward', 'backward'}
        Apply the forward projection if 'forward', otherwise the back
        projection
    ndim : {2, 3}
        Number of dimensions of the projector
    vol_id : int
        ASTRA ID of the volume data object
    sino_id : int
        ASTRA ID of the projection data object
    proj_id : int
        ASTRA ID of the projector
    impl : {'cpu', 'cuda'}
        Implementation of the projector

    Returns
    -------
    id : int
        ASTRA internal ID for the new algorithm structure
    """
    if direction not in ('forward', 'backward'):
        raise ValueError("`direction` '{}' not understood".format(direction))
    if ndim not in (2, 3):
        raise ValueError('{}-dimensional projectors not supported'
                         ''.format(ndim))
    if impl not in ('cpu', 'cuda'):
        raise ValueError("`impl` type '{}' not understood"
                         ''.format(impl))
    if ndim is 3 and impl is 'cpu':
        raise NotImplementedError(
            '3d algorithms for cpu is not supported by ASTRA')
    if proj_id is None and impl is 'cpu':
        raise ValueError("'cpu' implementation requires projector ID")

    algo_map = {'forward': {2: {'cpu': 'FP', 'cuda': 'FP_CUDA'},
                            3: {'cpu': None, 'cuda': 'FP3D_CUDA'}},
                'backward': {2: {'cpu': 'BP', 'cuda': 'BP_CUDA'},
                             3: {'cpu': None, 'cuda': 'BP3D_CUDA'}}}

    algo_cfg = {'type': algo_map[direction][ndim][impl],
                'ProjectorId': proj_id,
                'ProjectionDataId': sino_id}
    if direction is 'forward':
        algo_cfg['VolumeDataId'] = vol_id
    else:
        algo_cfg['ReconstructionDataId'] = vol_id

    # Create ASTRA algorithm object
    return astra.algorithm.create(algo_cfg)


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Step length computation for optimization schemes."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()

from abc import ABCMeta, abstractmethod
import numpy as np

from odl.util.utility import with_metaclass


__all__ = ('LineSearch', 'BacktrackingLineSearch', 'ConstantLineSearch')


class LineSearch(with_metaclass(ABCMeta, object)):

    """Abstract base class for line search step length methods."""

    @abstractmethod
    def __call__(self, x, direction, dir_derivative):
        """Calculate step length in direction.

        Parameters
        ----------
        x : `LinearSpaceElement`
            The current point
        direction : `LinearSpaceElement`
            Search direction in which the line search should be computed
        dir_derivative : float
            Directional derivative along the ``direction``

        Returns
        -------
        step : float
            Computed step length.
        """


class BacktrackingLineSearch(LineSearch):

    """Backtracking line search for step length calculation.

    This methods approximately finds the longest step length fulfilling
    the Armijo-Goldstein condition.

    The line search algorithm is described in [BV2004]_, page 464
    (`book available online
    <http://stanford.edu/~boyd/cvxbook/bv_cvxbook.pdf>`_) and
    [GNS2009]_, pages 378--379. See also
    `Backtracking_line_search
    <https://en.wikipedia.org/wiki/Backtracking_line_search>`_.
    """

    def __init__(self, function, tau=0.5, discount=0.01, max_num_iter=None,
                 estimate_step=False):
        """Initialize a new instance.

        Parameters
        ----------
        function : callable
            The cost function of the optimization problem to be solved.
        tau : float, optional
            The amount the step length is decreased in each iteration,
            as long as it does not fulfill the decrease condition.
            The step length is updated as ``step_length *= tau``.
        discount : float, optional
            The "discount factor" on ``step length * direction derivative``,
            yielding the threshold under which the function value must lie to
            be accepted (see the references).
        max_num_iter : int, optional
            Maximum number of iterations allowed each time the line
            search method is called. If not set, this number is calculated
            to allow a shortest step length of 10 times machine epsilon.
        estimate_step : float
            If the last step should be used as a estimate for the next step.
        """
        self.function = function
        self.tau = float(tau)
        self.discount = float(discount)
        self.estimate_step = bool(estimate_step)

        self.alpha = 1.0
        self.total_num_iter = 0
        # Use a default value that allows the shortest step to be < 0.0001
        # times the original step length
        if max_num_iter is None:
            # TODO: make space dependent
            eps = 10 * np.finfo(float).resolution
            self.max_num_iter = int(np.ceil(np.log(eps) / np.log(self.tau)))
        else:
            self.max_num_iter = int(max_num_iter)

    def __call__(self, x, direction, dir_derivative):
        """Calculate the optimal step length along a line.

        Parameters
        ----------
        x : `LinearSpaceElement`
            The current point
        direction : `LinearSpaceElement`
            Search direction in which the line search should be computed
        dir_derivative : float
            Directional derivative along the ``direction``

        Returns
        -------
        step : float
            The computed step length
        """
        fx = self.function(x)
        dir_derivative = float(dir_derivative)

        if dir_derivative == 0:
            raise ValueError('dir_derivative == 0, no descent can be found')
        if not self.estimate_step:
            alpha = 1.0
        else:
            alpha = self.alpha

        if dir_derivative > 0:
            # We need to move backwards if the direction is an increase
            # direction
            alpha *= -1

        if np.isnan(fx) or np.isinf(fx):
            raise ValueError('function returned invalid value {} in starting '
                             'point ({})'.format(fx, x))

        num_iter = 0
        while True:
            if num_iter > self.max_num_iter:
                raise ValueError('number of iterations exceeded maximum: {}, '
                                 'step length: {}, without finding a '
                                 'sufficient decrease'
                                 ''.format(self.max_num_iter, alpha))

            point = x + alpha * direction
            fval = self.function(point)

            if np.isnan(fval):
                # We do not want to compare against NaN below, and NaN should
                # indicate a user error.
                raise ValueError('function returned NaN in point '
                                 'point ({})'.format(point))

            expected_decrease = np.abs(alpha * dir_derivative * self.discount)
            if (not np.isinf(fval) and  # short circuit if fval is infite
                    fval <= fx - expected_decrease):
                # Stop iterating if the value decreases sufficiently.
                break

            num_iter += 1
            alpha *= self.tau

        assert fval < fx

        self.total_num_iter += num_iter
        self.alpha = np.abs(alpha)  # Store magnitude
        return alpha


class ConstantLineSearch(LineSearch):

    """Line search object that returns a constant step length."""

    def __init__(self, constant):
        """Initialize a new instance.

        Parameters
        ----------
        constant : float
            The constant step length
        """
        self.constant = float(constant)

    def __call__(self, x, direction, dir_derivative):
        """Calculate the step length at a point.

        Parameters
        ----------
        x : `LinearSpaceElement`
            The current point
        direction : `LinearSpaceElement`
            Search direction in which the line search should be computed
        dir_derivative : float
            Directional derivative along the ``direction``

        Returns
        -------
        step : float
            The constant step length
        """
        return self.constant


class BarzilaiBorweinStep(object):

    """Barzilai-Borwein method to compute a step length.

    Barzilai-Borwein method to compute a step length
    for gradient descent methods.

    The method is described in [BB1988]_ and [Ray1997]_.
    """

    def __init__(self, gradf, step0=0.0005):
        """Initialize a new instance.

        Parameters
        ----------
        gradf : `Operator`
            The gradient of the objective function at a point
        step0 : float, optional
            Initial step length parameter
        """
        self.gradf = gradf
        self.step0 = step0

    def __call__(self, x, x0):
        """Calculate the step length at a point.

        Parameters
        ----------
        x : `LinearSpaceElement`
            The current point
        x0 : `LinearSpaceElement`
            The previous point

        Returns
        -------
        step : float
            Computed step length.
        """
        if x == x0:
            return self.step0

        gradx = self.gradf(x)

        if gradx == self.gradf(x0):
            return self.step0

        errx = x - x0
        grad_diff = gradx - self.gradf(x0)
        recip_step = grad_diff.inner(errx) / errx.norm() ** 2
        return 1.0 / recip_step


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Operators defined for tensor fields."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import super

import numpy as np

from odl.operator import Operator
from odl.set import RealNumbers, ComplexNumbers, LinearSpace
from odl.space import ProductSpace


__all__ = ('PointwiseNorm', 'PointwiseInner')

_SUPPORTED_DIFF_METHODS = ('central', 'forward', 'backward')


class PointwiseTensorFieldOperator(Operator):

    """Abstract operator for point-wise tensor field manipulations.

    A point-wise operator acts on a space of vector or tensor fields,
    i.e. a power space ``X^d`` of a discretized function space ``X``.
    Its range is the power space ``X^k`` with a possibly different
    number ``k`` of components. For ``k == 1``, the base space
    ``X`` can be used instead.

    For example, if ``X`` is a `DiscreteLp` space, then
    ``ProductSpace(X, d)`` is a valid domain for any positive integer
    ``d``.

    Currently, only vector fields, i.e. one-dimensional products of
    ``X``, are supported.

    See Also
    --------
    ProductSpace
    """

    def __init__(self, domain, range, linear=False):
        """Initialize a new instance.

        Parameters
        ----------
        domain, range : {`ProductSpace`, `LinearSpace`}
            Spaces of vector fields between which the operator maps.
            They have to be either power spaces of the same base space
            ``X`` or the base space itself (only one of them).
            Empty product spaces are not allowed.
        linear : bool, optional
            If ``True``, assume that the operator is linear.
        """
        if isinstance(domain, ProductSpace):
            if not domain.is_power_space:
                raise TypeError('`domain` {!r} is not a power space'
                                ''.format(domain))

            if domain.size == 0:
                raise ValueError('`domain` is a product space of size 0')

            dom_base = domain[0]
        elif isinstance(domain, LinearSpace):
            dom_base = domain
        else:
            raise TypeError('`domain` {!r} is not a ProductSpace or '
                            'LinearSpace instance'.format(domain))

        if isinstance(range, ProductSpace):
            if not range.is_power_space:
                raise TypeError('`range` {!r} is not a power space'
                                ''.format(range))

            if range.size == 0:
                raise ValueError('`range` is a product space of size 0')

            ran_base = range[0]
        elif isinstance(range, LinearSpace):
            ran_base = range
        else:
            raise TypeError('`range` {!r} is not a ProductSpace or '
                            'LinearSpace instance'.format(range))

        if dom_base != ran_base:
            raise ValueError('`domain` and `range` have different base spaces '
                             '({!r} != {!r})'
                             ''.format(dom_base, ran_base))

        super().__init__(domain=domain, range=range, linear=linear)
        self.__base_space = dom_base

    @property
    def base_space(self):
        """Base space ``X`` of this operator's domain and range."""
        return self.__base_space


class PointwiseNorm(PointwiseTensorFieldOperator):

    """Take the point-wise norm of a vector field.

    This operator takes the (weighted) ``p``-norm

        ``||F(x)|| = [ sum_j( w_j * |F_j(x)|^p ) ]^(1/p)``

    for ``p`` finite and

        ``||F(x)|| = max_j( w_j * |F_j(x)| )``

    for ``p = inf``, where ``F`` is a vector field. This implies that
    the `Operator.domain` is a power space of a discretized function
    space. For example, if ``X`` is a `DiscreteLp` space, then
    ``ProductSpace(X, d)`` is a valid domain for any positive integer
    ``d``.
    """

    def __init__(self, vfspace, exponent=None, weight=None):
        """Initialize a new instance.

        Parameters
        ----------
        vfspace : `ProductSpace`
            Space of vector fields on which the operator acts.
            It has to be a product space of identical spaces, i.e. a
            power space.
        exponent : non-zero float, optional
            Exponent of the norm in each point. Values between
            0 and 1 are currently not supported due to numerical
            instability.
            Default: ``vfspace.exponent``
        weight : `array-like` or float, optional
            Weighting array or constant for the norm. If an array is
            given, its length must be equal to ``domain.size``, and
            all entries must be positive. A provided constant must be
            positive.
            By default, the weights are is taken from
            ``domain.weighting``. Note that this excludes unusual
            weightings with custom inner product, norm or dist.

        Examples
        --------
        We make a tiny vector field space in 2D and create the
        standard point-wise norm operator on that space. The operator
        maps a vector field to a scalar function:

        >>> spc = odl.uniform_discr([-1, -1], [1, 1], (1, 2))
        >>> vfspace = odl.ProductSpace(spc, 2)
        >>> pw_norm = PointwiseNorm(vfspace)
        >>> pw_norm.range == spc
        True

        Now we can calculate the 2-norm in each point:

        >>> x = vfspace.element([[[1, -4]],
        ...                      [[0, 3]]])
        >>> print(pw_norm(x))
        [[1.0, 5.0]]

        We can change the exponent either in the vector field space
        or in the operator directly:

        >>> vfspace = odl.ProductSpace(spc, 2, exponent=1)
        >>> pw_norm = PointwiseNorm(vfspace)
        >>> print(pw_norm(x))
        [[1.0, 7.0]]
        >>> vfspace = odl.ProductSpace(spc, 2)
        >>> pw_norm = PointwiseNorm(vfspace, exponent=1)
        >>> print(pw_norm(x))
        [[1.0, 7.0]]
        """
        if not isinstance(vfspace, ProductSpace):
            raise TypeError('`vfspace` {!r} is not a ProductSpace '
                            'instance'.format(vfspace))
        super().__init__(domain=vfspace, range=vfspace[0], linear=False)

        # Need to check for product space shape once higher order tensors
        # are implemented

        if exponent is None:
            if self.domain.exponent is None:
                raise ValueError('cannot determine `exponent` from {}'
                                 ''.format(self.domain))
            self._exponent = self.domain.exponent
        elif exponent < 1:
            raise ValueError('`exponent` smaller than 1 not allowed')
        else:
            self._exponent = float(exponent)

        # Handle weighting, including sanity checks
        if weight is None:
            # TODO: find a more robust way of getting the weighs as a vector
            if hasattr(self.domain.weighting, 'vector'):
                self._weights = self.domain.weighting.vector
            elif hasattr(self.domain.weighting, 'const'):
                self._weights = (self.domain.weighting.const *
                                 np.ones(len(self.domain)))
            else:
                raise ValueError('weighting scheme {!r} of the domain does '
                                 'not define a weighting vector or constant'
                                 ''.format(self.domain.weighting))
        elif np.isscalar(weight):
            if weight <= 0:
                raise ValueError('weighting constant must be positive, got '
                                 '{}'.format(weight))
            self._weights = float(weight) * np.ones(self.domain.size)
        else:
            self._weights = np.asarray(weight, dtype='float64')
            if (not np.all(self.weights > 0) or
                    not np.all(np.isfinite(self.weights))):
                raise ValueError('weighting array {} contains invalid '
                                 'entries'.format(weight))
        self._is_weighted = not np.array_equiv(self.weights, 1.0)

    @property
    def exponent(self):
        """Exponent ``p`` of this norm."""
        return self._exponent

    @property
    def weights(self):
        """Weighting vector of this norm."""
        return self._weights

    @property
    def is_weighted(self):
        """``True`` if weighting is not 1 or all ones."""
        return self._is_weighted

    def _call(self, f, out):
        """Implement ``self(f, out)``."""
        if self.exponent == 1.0:
            self._call_vecfield_1(f, out)
        elif self.exponent == float('inf'):
            self._call_vecfield_inf(f, out)
        else:
            self._call_vecfield_p(f, out)

    def _call_vecfield_1(self, vf, out):
        """Implement ``self(vf, out)`` for exponent 1."""
        vf[0].ufunc.absolute(out=out)
        if self.is_weighted:
            out *= self.weights[0]

        if self.domain.size == 1:
            return

        tmp = self.range.element()
        for fi, wi in zip(vf[1:], self.weights[1:]):
            fi.ufunc.absolute(out=tmp)
            if self.is_weighted:
                tmp *= wi
            out += tmp

    def _call_vecfield_inf(self, vf, out):
        """Implement ``self(vf, out)`` for exponent ``inf``."""
        vf[0].ufunc.absolute(out=out)
        if self.is_weighted:
            out *= self.weights[0]

        if self.domain.size == 1:
            return

        tmp = self.range.element()
        for vfi, wi in zip(vf[1:], self.weights[1:]):
            vfi.ufunc.absolute(out=tmp)
            if self.is_weighted:
                tmp *= wi
            out.ufunc.maximum(tmp, out=out)

    def _call_vecfield_p(self, vf, out):
        """Implement ``self(vf, out)`` for exponent 1 < p < ``inf``."""
        # Optimization for 1 component - just absolute value (maybe weighted)
        if self.domain.size == 1:
            vf[0].ufunc.absolute(out=out)
            if self.is_weighted:
                out *= self.weights[0] ** (1 / self.exponent)
            return

        # Initialize out, avoiding one copy
        self._abs_pow_ufunc(vf[0], out=out)
        if self.is_weighted:
            out *= self.weights[0]

        tmp = self.range.element()
        for fi, wi in zip(vf[1:], self.weights[1:]):
            self._abs_pow_ufunc(fi, out=tmp)
            if self.is_weighted:
                tmp *= wi
            out += tmp

        out.ufunc.power(1 / self.exponent, out=out)

    def _abs_pow_ufunc(self, fi, out):
        """Compute |F_i(x)|^p point-wise and write to ``out``."""
        # Optimization for a very common case
        if self.exponent == 2.0 and self.base_space.field == RealNumbers():
            fi.multiply(fi, out=out)
        else:
            fi.ufunc.absolute(out=out)
            out.ufunc.power(self.exponent, out=out)

    def derivative(self, vf):
        """Derivative of the point-wise norm operator at ``vf``.

        The derivative at ``F`` of the point-wise norm operator ``N``
        with finite exponent ``p`` and weights ``w`` is the pointwise
        inner product with the vector field

            ``x --> N(F)(x)^(1-p) * [ F_j(x) * |F_j(x)|^(p-2) ]_j``.

        Note that this is not well-defined for ``F = 0``. If ``p < 2``,
        any zero component will result in a singularity.

        Parameters
        ----------
        vf : `domain` `element-like`
            Vector field ``F`` at which to evaluate the derivative.

        Returns
        -------
        deriv : `PointwiseInner`
            Derivative operator at the given point ``vf``.

        Raises
        ------
        NotImplementedError
            * if the vector field space is complex, since the derivative
              is not linear in that case
            * if the exponent is ``inf``
        """
        if self.domain.field == ComplexNumbers():
            raise NotImplementedError('operator not Frechet-differentiable '
                                      'on a complex space')

        if self.exponent == float('inf'):
            raise NotImplementedError('operator not Frechet-differentiable '
                                      'for exponent = inf')

        vf = self.domain.element(vf)
        vf_pwnorm_fac = self(vf)
        vf_pwnorm_fac **= (self.exponent - 1)

        inner_vf = vf.copy()

        for gi in inner_vf:
            gi /= vf_pwnorm_fac * gi ** (self.exponent - 2)

        return PointwiseInner(self.domain, inner_vf, weight=self.weights)


class PointwiseInnerBase(PointwiseTensorFieldOperator):
    """Base class for `PointwiseInner` and `PointwiseInnerAdjoint`.

    Implemented to allow code reuse between the classes.
    """

    def __init__(self, adjoint, vfspace, vecfield, weight=None):
        """Initialize a new instance.

        All parameters are given according to the specifics of the "usual"
        operator. The ``adjoint`` parameter is used to control conversions
        for the inverse transform.

        Parameters
        ----------
        adjoint : bool
            ``True`` if the operator should be the adjoint, ``False``
            otherwise.
        vfspace : `ProductSpace`
            Space of vector fields on which the operator acts.
            It has to be a product space of identical spaces, i.e. a
            power space.
        vecfield : ``vfspace`` `element-like`
            Vector field with which to calculate the point-wise inner
            product of an input vector field
        weight : `array-like` or float, optional
            Weighting array or constant for the norm. If an array is
            given, its length must be equal to ``domain.size``, and
            all entries must be positive. A provided constant must be
            positive.
            By default, the weights are is taken from
            ``domain.weighting``. Note that this excludes unusual
            weightings with custom inner product, norm or dist.
        """
        if not isinstance(vfspace, ProductSpace):
            raise TypeError('`vfsoace` {!r} is not a ProductSpace '
                            'instance'.format(vfspace))
        if adjoint:
            super().__init__(domain=vfspace[0], range=vfspace, linear=True)
        else:
            super().__init__(domain=vfspace, range=vfspace[0], linear=True)

        # Bail out if the space is complex but we cannot take the complex
        # conjugate.
        if (vfspace.field == ComplexNumbers() and
                not hasattr(self.base_space.element_type, 'conj')):
            raise NotImplementedError(
                'base space element type {!r} does not implement conj() '
                'method required for complex inner products'
                ''.format(self.base_space.element_type))

        self._vecfield = vfspace.element(vecfield)

        # Handle weighting, including sanity checks
        if weight is None:
            if hasattr(vfspace.weighting, 'vector'):
                self._weights = vfspace.weighting.vector
            elif hasattr(vfspace.weighting, 'const'):
                self._weights = (vfspace.weighting.const *
                                 np.ones(len(vfspace)))
            else:
                raise ValueError('weighting scheme {!r} of the domain does '
                                 'not define a weighting vector or constant'
                                 ''.format(vfspace.weighting))
        elif np.isscalar(weight):
            if weight <= 0:
                raise ValueError('weighting constant must be positive, got '
                                 '{}'.format(weight))
            self._weights = float(weight) * np.ones(vfspace.size)
        else:
            self._weights = np.asarray(weight, dtype='float64')
            if (not np.all(self.weights > 0) or
                    not np.all(np.isfinite(self.weights))):
                raise ValueError('weighting array {} contains invalid '
                                 'entries'.format(weight))
        self._is_weighted = not np.array_equiv(self.weights, 1.0)

    @property
    def vecfield(self):
        """Fixed vector field ``G`` of this inner product."""
        return self._vecfield

    @property
    def weights(self):
        """Weighting vector of this norm."""
        return self._weights

    @property
    def is_weighted(self):
        """``True`` if weighting is not 1 or all ones."""
        return self._is_weighted

    @property
    def adjoint(self):
        """Adjoint operator."""
        raise NotImplementedError('abstract method')


class PointwiseInner(PointwiseInnerBase):

    """Take the point-wise inner product with a given vector field.

    This operator takes the (weighted) inner product

        ``<F(x), G(x)> = sum_j ( w_j * F_j(x) * conj(G_j(x)) )``

    for a given vector field ``G``, where ``F`` is the vector field
    acting as a variable to this operator.

    This implies that the `Operator.domain` is a power space of a
    discretized function space. For example, if ``X`` is a `DiscreteLp`
    space, then ``ProductSpace(X, d)`` is a valid domain for any
    positive integer ``d``.
    """

    def __init__(self, vfspace, vecfield, weight=None):
        """Initialize a new instance.

        Parameters
        ----------
        vfspace : `ProductSpace`
            Space of vector fields on which the operator acts.
            It has to be a product space of identical spaces, i.e. a
            power space.
        vecfield : ``vfspace`` `element-like`
            Vector field with which to calculate the point-wise inner
            product of an input vector field
        weight : `array-like` or float, optional
            Weighting array or constant for the norm. If an array is
            given, its length must be equal to ``domain.size``, and
            all entries must be positive. A provided constant must be
            positive.
            By default, the weights are is taken from
            ``domain.weighting``. Note that this excludes unusual
            weightings with custom inner product, norm or dist.

        Examples
        --------
        We make a tiny vector field space in 2D and create the
        point-wise inner product operator with a fixed vector field.
        The operator maps a vector field to a scalar function:

        >>> spc = odl.uniform_discr([-1, -1], [1, 1], (1, 2))
        >>> vfspace = odl.ProductSpace(spc, 2)
        >>> fixed_vf = np.array([[[0, 1]],
        ...                      [[1, -1]]])
        >>> pw_inner = PointwiseInner(vfspace, fixed_vf)
        >>> pw_inner.range == spc
        True

        Now we can calculate the inner product in each point:

        >>> x = vfspace.element([[[1, -4]],
        ...                      [[0, 3]]])
        >>> print(pw_inner(x))
        [[0.0, -7.0]]
        """
        super().__init__(adjoint=False, vfspace=vfspace, vecfield=vecfield,
                         weight=weight)

    @property
    def vecfield(self):
        """Fixed vector field ``G`` of this inner product."""
        return self._vecfield

    @property
    def weights(self):
        """Weighting vector of this norm."""
        return self._weights

    @property
    def is_weighted(self):
        """``True`` if weighting is not 1 or all ones."""
        return self._is_weighted

    def _call(self, vf, out):
        """Implement ``self(vf, out)``."""
        if self.domain.field == ComplexNumbers():
            vf[0].multiply(self._vecfield[0].conj(), out=out)
        else:
            vf[0].multiply(self._vecfield[0], out=out)

        if self.is_weighted:
            out *= self.weights[0]

        if self.domain.size == 1:
            return

        tmp = self.range.element()
        for vfi, gi, wi in zip(vf[1:], self.vecfield[1:],
                               self.weights[1:]):

            if self.domain.field == ComplexNumbers():
                vfi.multiply(gi.conj(), out=tmp)
            else:
                vfi.multiply(gi, out=tmp)

            if self.is_weighted:
                tmp *= wi
            out += tmp

    @property
    def adjoint(self):
        """Adjoint of this operator.

        Returns
        -------
        adjoint : `PointwiseInnerAdjoint`
        """
        return PointwiseInnerAdjoint(
            sspace=self.base_space, vecfield=self.vecfield,
            vfspace=self.domain, weight=self.weights)


class PointwiseInnerAdjoint(PointwiseInnerBase):

    """Adjoint of the point-wise inner product operator.

    The adjoint of the inner product operator is a mapping

        ``A^* : X --> X^d``.

    If the vector field space ``X^d`` is weighted by a vector ``v``,
    the adjoint, applied to a function ``h`` from ``X`` is the vector
    field

        ``x --> h(x) * (w / v) * G(x)``,

    where ``G`` and ``w`` are the vector field and weighting from the
    inner product operator, resp., and all multiplications are understood
    component-wise.
    """

    def __init__(self, sspace, vecfield, vfspace=None, weight=None):
        """Initialize a new instance.

        Parameters
        ----------
        sspace : `LinearSpace`
            "Scalar" space on which the operator acts
        vecfield : `range` `element-like`
            Vector field of the point-wise inner product operator
        vfspace : `ProductSpace`, optional
            Space of vector fields to which the operator maps. It must
            be a power space with ``sspace`` as base space.
            This option is intended to enforce an operator range
            with a certain weighting.
            Default: ``ProductSpace(space, len(vecfield), weight=weight)``
        weight : `array-like` or float, optional
            Weighting array or constant of the inner product operator.
            If an array is given, its length must be equal to
            ``len(vecfield)``, and all entries must be positive. A
            provided constant must be positive.
            By default, the weights are is taken from
            ``range.weighting`` if applicable. Note that this excludes
            unusual weightings with custom inner product, norm or dist.
        """
        if vfspace is None:
            vfspace = ProductSpace(sspace, len(vecfield), weight=weight)
        else:
            if not isinstance(vfspace, ProductSpace):
                raise TypeError('`vfspace` {!r} is not a '
                                'ProductSpace instance'.format(vfspace))
            if vfspace[0] != sspace:
                raise ValueError('base space of the range is different from '
                                 'the given scalar space ({!r} != {!r})'
                                 ''.format(vfspace[0], sspace))
        super().__init__(adjoint=True, vfspace=vfspace, vecfield=vecfield,
                         weight=weight)

        # Get weighting from range
        if hasattr(self.range.weighting, 'vector'):
            self._ran_weights = self.range.weighting.vector
        elif hasattr(self.range.weighting, 'const'):
            self._ran_weights = (self.range.weighting.const *
                                 np.ones(len(self.range)))
        else:
            raise ValueError('weighting scheme {!r} of the range does '
                             'not define a weighting vector or constant'
                             ''.format(self.range.weighting))

    def _call(self, f, out):
        """Implement ``self(vf, out)``."""
        for vfi, oi, ran_wi, dom_wi in zip(self.vecfield, out,
                                           self._ran_weights, self.weights):
            vfi.multiply(f, out=oi)
            if not np.isclose(ran_wi, dom_wi):
                oi *= dom_wi / ran_wi

    @property
    def adjoint(self):
        """Adjoint of this operator.

        Returns
        -------
        adjoint : `PointwiseInner`
        """
        return PointwiseInner(vfspace=self.range, vecfield=self.vecfield,
                              weight=self.weights)


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""UFuncs for ODL vectors.

These functions are internal and should only be used as methods on
`NtuplesBaseVector` type spaces.

See `numpy.ufuncs
<http://docs.scipy.org/doc/numpy/reference/ufuncs.html>`_
for more information.

Notes
-----
The default implementation of these methods make heavy use of the
``NtuplesBaseVector.__array__`` to extract a `numpy.ndarray` from the vector,
and then apply a ufunc to it. Afterwards, ``NtuplesBaseVector.__array_wrap__``
is used to re-wrap the data into the appropriate space.
"""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()

import numpy as np
import re


__all__ = ('NtuplesBaseUFuncs', 'NumpyNtuplesUFuncs',
           'DiscreteLpUFuncs', 'ProductSpaceUFuncs')


# Some are ignored since they don't cooperate with dtypes, needs fix
RAW_UFUNCS = ['absolute', 'add', 'arccos', 'arccosh', 'arcsin', 'arcsinh',
              'arctan', 'arctan2', 'arctanh', 'bitwise_and', 'bitwise_or',
              'bitwise_xor', 'ceil', 'conj', 'copysign', 'cos', 'cosh',
              'deg2rad', 'divide', 'equal', 'exp', 'exp2', 'expm1', 'floor',
              'floor_divide', 'fmax', 'fmin', 'fmod', 'fmod', 'greater',
              'greater_equal', 'hypot', 'invert', 'isfinite', 'isinf', 'isnan',
              'left_shift', 'less', 'less_equal', 'log', 'log10', 'log1p',
              'log2', 'logaddexp', 'logaddexp2', 'logical_and', 'logical_not',
              'logical_or', 'logical_xor', 'maximum', 'minimum', 'mod', 'modf',
              'multiply', 'negative', 'not_equal', 'power',
              'rad2deg', 'reciprocal', 'remainder', 'right_shift', 'rint',
              'sign', 'signbit', 'sin', 'sinh', 'sqrt', 'square', 'subtract',
              'tan', 'tanh', 'true_divide', 'trunc']
# ,'isreal', 'iscomplex', 'ldexp', 'frexp'

# Add some standardized information
UFUNCS = []
for name in RAW_UFUNCS:
    ufunc = getattr(np, name)
    n_in, n_out = ufunc.nin, ufunc.nout
    descr = ufunc.__doc__.splitlines()[2]
    # Numpy occasionally uses single tics for doc, we only use them for links
    descr = re.sub('`+', '``', descr)
    doc = descr + """

See Also
--------
numpy.{}
""".format(name)
    UFUNCS.append((name, n_in, n_out, doc))

RAW_REDUCTIONS = [('sum', 'sum', 'Sum of array elements.'),
                  ('prod', 'prod', 'Product of array elements.'),
                  ('min', 'amin', 'Minimum value in array.'),
                  ('max', 'amax', 'Maximum value in array.')]

REDUCTIONS = []
for name, numpyname, descr in RAW_REDUCTIONS:
    doc = descr + """

See Also
--------
numpy.{}
""".format(numpyname)
    REDUCTIONS += [(name, doc)]


# Wrap all numpy ufuncs

def wrap_ufunc_base(name, n_in, n_out, doc):
    """Add ufunc methods to `NtuplesBaseUFuncs`."""
    wrapped = getattr(np, name)
    if n_in == 1:
        if n_out == 0:
            def wrapper(self):
                return wrapped(self.vector)

        elif n_out == 1:
            def wrapper(self, out=None):
                if out is None:
                    out = self.vector.space.element()

                out[:] = wrapped(self.vector)
                return out

        elif n_out == 2:
            def wrapper(self, out1=None, out2=None):
                if out1 is None:
                    out1 = self.vector.space.element()
                if out2 is None:
                    out2 = self.vector.space.element()

                [y1, y2] = wrapped(self.vector)
                out1[:] = y1
                out2[:] = y2
                return out1, out2

        else:
            raise NotImplementedError

    elif n_in == 2:
        if n_out == 1:
            def wrapper(self, x2, out=None):
                if out is None:
                    return wrapped(self.vector, x2)
                else:
                    out[:] = wrapped(self.vector, x2)
                    return out

        else:
            raise NotImplementedError
    else:
        raise NotImplementedError

    wrapper.__name__ = name
    wrapper.__doc__ = doc
    return wrapper


# Wrap reductions
def wrap_reduction_base(name, doc):
    """Add ufunc methods to `NtuplesBaseUFuncs`."""
    wrapped = getattr(np, name)

    def wrapper(self):
        return wrapped(self.vector)

    wrapper.__name__ = name
    wrapper.__doc__ = doc
    return wrapper


class NtuplesBaseUFuncs(object):

    """UFuncs for `NtuplesBaseVector` objects.

    Internal object, should not be created except in `NtuplesBaseVector`.
    """

    def __init__(self, vector):
        """Create ufunc wrapper for vector."""
        self.vector = vector


# Add ufunc methods to UFunc class
for name, n_in, n_out, doc in UFUNCS:
    method = wrap_ufunc_base(name, n_in, n_out, doc)
    setattr(NtuplesBaseUFuncs, name, method)

# Add reduction methods to UFunc class
for name, doc in REDUCTIONS:
    method = wrap_reduction_base(name, doc)
    setattr(NtuplesBaseUFuncs, name, method)


# Optimized implementation of ufuncs since we can use the out parameter
# as well as the data parameter to avoid one call to asarray() when using an
# NumpyNtuplesVector
def wrap_ufunc_ntuples(name, n_in, n_out, doc):
    """Add ufunc methods to `NumpyNtuplesUFuncs`."""

    # Get method from numpy
    wrapped = getattr(np, name)
    if n_in == 1:
        if n_out == 0:
            def wrapper(self):
                return wrapped(self.vector)

        elif n_out == 1:
            def wrapper(self, out=None):
                if out is None:
                    out = self.vector.space.element()
                wrapped(self.vector, out.data)
                return out

        elif n_out == 2:
            def wrapper(self, out1=None, out2=None):
                if out1 is None:
                    out1 = self.vector.space.element()
                if out2 is None:
                    out2 = self.vector.space.element()

                y1, y2 = wrapped(self.vector, out1.data, out2.data)
                return out1, out2

        else:
            raise NotImplementedError

    elif n_in == 2:
        if n_out == 1:
            def wrapper(self, x2, out=None):
                if out is None:
                    out = self.vector.space.element()

                wrapped(self.vector, x2, out.data)
                return out

        else:
            raise NotImplementedError
    else:
        raise NotImplementedError

    wrapper.__name__ = name
    wrapper.__doc__ = doc
    return wrapper


class NumpyNtuplesUFuncs(NtuplesBaseUFuncs):

    """UFuncs for `NumpyNtuplesVector` objects.

    Internal object, should not be created except in `NumpyNtuplesVector`.
    """


# Add ufunc methods to UFunc class
for name, n_in, n_out, doc in UFUNCS:
    method = wrap_ufunc_ntuples(name, n_in, n_out, doc)
    setattr(NumpyNtuplesUFuncs, name, method)


# Optimized implementation of ufuncs since we can use the out parameter
# as well as the data parameter to avoid one call to asarray() when using a
# NumpyNtuplesVector
def wrap_ufunc_discretelp(name, n_in, n_out, doc):
    """Add ufunc methods to `DiscreteLpUFuncs`."""

    if n_in == 1:
        if n_out == 0:
            def wrapper(self):
                method = getattr(self.vector.ntuple.ufunc, name)
                return self.vector.space.element(method())

        elif n_out == 1:
            def wrapper(self, out=None):
                method = getattr(self.vector.ntuple.ufunc, name)
                if out is None:
                    return self.vector.space.element(method())
                else:
                    method(out=out.ntuple)
                    return out

        elif n_out == 2:
            def wrapper(self, out1=None, out2=None):
                method = getattr(self.vector.ntuple.ufunc, name)
                if out1 is None:
                    out1 = self.vector.space.element()
                if out2 is None:
                    out2 = self.vector.space.element()

                y1, y2 = method(out1.ntuple, out2.ntuple)
                return out1, out2

        else:
            raise NotImplementedError

    elif n_in == 2:
        if n_out == 1:
            def wrapper(self, x2, out=None):
                if x2 in self.vector.space:
                    x2 = x2.ntuple

                try:
                    # Try to reshape to linear data
                    x2 = x2.reshape(self.vector.size,
                                    order=self.vector.space.order)
                except AttributeError:
                    pass

                method = getattr(self.vector.ntuple.ufunc, name)
                if out is None:
                    return self.vector.space.element(method(x2))
                else:
                    method(x2, out.ntuple)
                    return out

        else:
            raise NotImplementedError
    else:
        raise NotImplementedError

    wrapper.__name__ = name
    wrapper.__doc__ = doc
    return wrapper


def wrap_reduction_discretelp(name, doc):
    def wrapper(self):
        method = getattr(self.vector.ntuple.ufunc, name)
        return method()

    wrapper.__name__ = name
    wrapper.__doc__ = doc
    return wrapper


class DiscreteLpUFuncs(NtuplesBaseUFuncs):

    """UFuncs for `DiscreteLpElement` objects.

    Internal object, should not be created except in `DiscreteLpElement`.
    """


# Add ufunc methods to UFunc class
for name, n_in, n_out, doc in UFUNCS:
    method = wrap_ufunc_discretelp(name, n_in, n_out, doc)
    setattr(DiscreteLpUFuncs, name, method)

for name, doc in REDUCTIONS:
    method = wrap_reduction_discretelp(name, doc)
    setattr(DiscreteLpUFuncs, name, method)


# Ufuncs for product space elements
def wrap_ufunc_productspace(name, n_in, n_out, doc):
    """Add ufunc methods to `ProductSpaceElement`."""

    if n_in == 1:
        if n_out == 0:
            def wrapper(self):
                result = [getattr(x.ufunc, name)() for x in self.vector]
                return self.vector.space.element(result)

        elif n_out == 1:
            def wrapper(self, out=None):
                if out is None:
                    result = [getattr(x.ufunc, name)() for x in self.vector]
                    return self.vector.space.element(result)
                else:
                    for x, out_x in zip(self.vector, out):
                        getattr(x.ufunc, name)(out=out_x)
                    return out

        elif n_out == 2:
            def wrapper(self, out1=None, out2=None):
                if out1 is None:
                    out1 = self.vector.space.element()
                if out2 is None:
                    out2 = self.vector.space.element()
                for x, out1_x, out2_x in zip(self.vector, out1, out2):
                    getattr(x.ufunc, name)(out1=out1_x, out2=out2_x)
                return out1, out2

        else:
            raise NotImplementedError

    elif n_in == 2:
        if n_out == 1:
            def wrapper(self, x2, out=None):
                if x2 in self.vector.space:
                    if out is None:
                        result = [getattr(x.ufunc, name)(x2p)
                                  for x, x2p in zip(self.vector, x2)]
                        return self.vector.space.element(result)
                    else:
                        for x, x2p, outp in zip(self.vector, x2, out):
                            getattr(x.ufunc, name)(x2p, out=outp)
                        return out
                else:
                    if out is None:
                        result = [getattr(x.ufunc, name)(x2)
                                  for x in self.vector]
                        return self.vector.space.element(result)
                    else:
                        for x, outp in zip(self.vector, out):
                            getattr(x.ufunc, name)(x2, out=outp)
                        return out

        else:
            raise NotImplementedError
    else:
        raise NotImplementedError

    wrapper.__name__ = name
    wrapper.__doc__ = doc
    return wrapper


def wrap_reduction_productspace(name, doc):
    """Add reduction methods to `ProductSpaceElement`."""
    def wrapper(self):
        results = [getattr(x.ufunc, name)() for x in self.vector]
        return getattr(np, name)(results)

    wrapper.__name__ = name
    wrapper.__doc__ = doc
    return wrapper


class ProductSpaceUFuncs(object):

    """UFuncs for `ProductSpaceElement` objects.

    Internal object, should not be created except in `ProductSpaceElement`.
    """
    def __init__(self, vector):
        """Create ufunc wrapper for vector."""
        self.vector = vector


# Add ufunc methods to UFunc class
for name, n_in, n_out, doc in UFUNCS:
    method = wrap_ufunc_productspace(name, n_in, n_out, doc)
    setattr(ProductSpaceUFuncs, name, method)


# Add reduction methods to UFunc class
for name, doc in REDUCTIONS:
    method = wrap_reduction_productspace(name, doc)
    setattr(ProductSpaceUFuncs, name, method)



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Domains for continuous functions. """

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from builtins import super, zip
from future import standard_library
from future.utils import raise_from
standard_library.install_aliases()

import numpy as np

from odl.set.sets import Set
from odl.util.utility import array1d_repr
from odl.util.vectorization import (
    is_valid_input_array, is_valid_input_meshgrid)


__all__ = ('IntervalProd',)


class IntervalProd(Set):

    """An n-dimensional rectangular box.

    An interval product is a Cartesian product of n intervals, i.e. an
    n-dimensional rectangular box aligned with the coordinate axes
    as a subset of the n-dimensional Euclidean space.

    `IntervalProd` objects are immutable, hence all manipulation methods
    return a new instance.
    """

    def __init__(self, min_pt, max_pt):
        """Initialize a new instance.

        Parameters
        ----------
        min_pt, max_pt : float or `array-like`
            Vectors of lower/upper ends of the intervals in the product.

        Examples
        --------
        >>> min_pt, max_pt = [-1, 2.5, 70, 80], [-0.5, 10, 75, 90]
        >>> rbox = odl.IntervalProd(min_pt, max_pt)
        >>> rbox
        IntervalProd([-1.0, 2.5, 70.0, 80.0], [-0.5, 10.0, 75.0, 90.0])
        """
        self.__min_pt = np.atleast_1d(min_pt).astype('float64')
        self.__max_pt = np.atleast_1d(max_pt).astype('float64')

        if self.min_pt.ndim > 1:
            raise ValueError('`min_pt` must be 1-dimensional, got an array '
                             'with {} axes'.format(self.min_pt.ndim))
        if self.max_pt.ndim > 1:
            raise ValueError('`max_pt` must be 1-dimensional, got an array '
                             'with {} axes'.format(self.max_pt.ndim))
        if len(self.min_pt) != len(self.max_pt):
            raise ValueError('`min_pt` and `max_pt` have different lengths '
                             '({} != {})'
                             ''.format(len(self.min_pt), len(self.max_pt)))

        for axis, (xmin, xmax) in enumerate(zip(self.min_pt, self.max_pt)):
            if xmax < xmin:
                raise ValueError('in axis {}: upper end smaller than lower '
                                 'end ({} < {})'.format(axis, xmax, xmin))

        self.__ideg = np.where(self.min_pt == self.max_pt)[0]
        self.__inondeg = np.where(self.min_pt != self.max_pt)[0]
        super().__init__()

    @property
    def min_pt(self):
        """Left interval boundaries of this interval product."""
        return self.__min_pt

    @property
    def max_pt(self):
        """Right interval boundaries of this interval product."""
        return self.__max_pt

    @property
    def ndim(self):
        """Number of intervals in the product."""
        return len(self.min_pt)

    @property
    def true_ndim(self):
        """Number of non-degenerate (positive-length) intervals."""
        return len(self.inondeg)

    @property
    def volume(self):
        """`ndim`-dimensional volume of this interval product."""
        return self.measure(ndim=self.ndim)

    @property
    def length(self):
        """Length of this interval (valid for ``ndim == 1``)."""
        if self.ndim != 1:
            raise NotImplementedError('length not defined if `ndim` != 1')
        return self.volume

    @property
    def area(self):
        """Area of this rectangle (valid if ``ndim == 2``)."""
        if self.ndim != 2:
            raise NotImplementedError('area not defined if `ndim` != 2')
        return self.volume

    @property
    def mid_pt(self):
        """Midpoint of this interval product."""
        midp = (self.max_pt + self.min_pt) / 2.
        midp[self.ideg] = self.min_pt[self.ideg]
        return midp

    @property
    def ideg(self):
        """Indices of the degenerate dimensions."""
        return self.__ideg

    @property
    def inondeg(self):
        """Indices of the non-degenerate dimensions."""
        return self.__inondeg

    def min(self):
        """Return the minimum point of this interval product."""
        return self.min_pt

    def max(self):
        """Return the maximum point of this interval product."""
        return self.max_pt

    def extent(self):
        """Return the vector of interval lengths per axis."""
        return self.max() - self.min()

    def element(self, inp=None):
        """Return an element of this interval product.

        Parameters
        ----------
        inp : float or `array-like`, optional
            Point to be cast to an element.

        Returns
        -------
        element : `numpy.ndarray` or float
            Array (`ndim` > 1) or float version of ``inp`` if provided,
            otherwise ``self.mid_pt``.

        Examples
        --------
        >>> interv = IntervalProd(0, 1)
        >>> interv.element(0.5)
        0.5
        """
        if inp is None:
            return self.mid_pt
        elif inp in self:
            if self.ndim == 1:
                return float(inp)
            else:
                return np.asarray(inp)
        else:
            raise TypeError('`inp` {!r} is not a valid element of {!r}'
                            ''.format(inp, self))

    def approx_equals(self, other, atol):
        """Return ``True`` if ``other`` is equal to this set up to ``atol``.

        Parameters
        ----------
        other :
            Object to be tested.
        atol : float
            Maximum allowed difference in maximum norm between the
            interval endpoints.

        Examples
        --------
        >>> rbox1 = IntervalProd(0, 0.5)
        >>> rbox2 = IntervalProd(0, np.sqrt(0.5)**2)
        >>> rbox1.approx_equals(rbox2, atol=0)  # Numerical error
        False
        >>> rbox1.approx_equals(rbox2, atol=1e-15)
        True
        """
        if other is self:
            return True
        elif not isinstance(other, IntervalProd):
            return False

        return (np.allclose(self.min_pt, other.min_pt, atol=atol, rtol=0.0) and
                np.allclose(self.max_pt, other.max_pt, atol=atol, rtol=0.0))

    def __eq__(self, other):
        """Return ``self == other``."""
        # Implemented separately for performance reasons
        if other is self:
            return True
        elif not isinstance(other, IntervalProd):
            return False

        return (np.all(self.min_pt == other.min_pt) and
                np.all(self.max_pt == other.max_pt))

    def approx_contains(self, point, atol):
        """Return ``True`` if ``point`` is "almost" contained in this set.

        Parameters
        ----------
        point : `array-like` or float
            Point to be tested. Its length must be equal to `ndim`.
            In the 1d case, ``point`` can be given as a float.
        atol : float
            Maximum allowed distance in maximum norm from ``point``
            to ``self``.

        Examples
        --------
        >>> min_pt, max_pt = [-1, 0, 2], [-0.5, 0, 3]
        >>> rbox = IntervalProd(min_pt, max_pt)
        >>> # Numerical error
        >>> rbox.approx_contains([-1 + np.sqrt(0.5)**2, 0., 2.9], atol=0)
        False
        >>> rbox.approx_contains([-1 + np.sqrt(0.5)**2, 0., 2.9], atol=1e-9)
        True
        """
        try:
            # Duck-typed check of type
            point = np.array(point, dtype=np.float, copy=False, ndmin=1)
        except (ValueError, TypeError):
            return False

        if point.shape != (self.ndim,):
            return False

        return self.dist(point, exponent=np.inf) <= atol

    def __contains__(self, other):
        """Return ``other in self``.

        Examples
        --------
        >>> interv = IntervalProd(0, 1)
        >>> 0.5 in interv
        True
        >>> 2 in interv
        False
        >>> 'string' in interv
        False
        """
        try:
            # Duck-typed check of type
            point = np.array(other, dtype=np.float, copy=False, ndmin=1)
        except (ValueError, TypeError):
            return False

        if point.shape != (self.ndim,):
            return False
        return (self.min_pt <= point).all() and (point <= self.max_pt).all()

    def contains_set(self, other, atol=0.0):
        """Return ``True`` if ``other`` is (almost) contained in this set.

        Parameters
        ----------
        other : `Set`
            Set to be tested.
        atol : float, optional
            Maximum allowed distance in maximum norm from ``other``
            to ``self``.

        Raises
        ------
        AttributeError
            if ``other`` does not have both ``min`` and ``max`` methods.

        Examples
        --------
        >>> min_pt1, max_pt1 = [-1, 0, 2], [-0.5, 0, 3]
        >>> rbox1 = IntervalProd(min_pt1, max_pt1)
        >>> min_pt2, max_pt2 = [-0.6, 0, 2.1], [-0.5, 0, 2.5]
        >>> rbox2 = IntervalProd(min_pt2, max_pt2)
        >>> rbox1.contains_set(rbox2)
        True
        >>> rbox2.contains_set(rbox1)
        False
        """
        if self is other:
            return True

        try:
            return (self.approx_contains(other.min(), atol) and
                    self.approx_contains(other.max(), atol))
        except AttributeError as err:
            raise_from(
                AttributeError('cannot test {!r} without `min` and `max` '
                               'methods'.format(other)), err)

    def contains_all(self, other, atol=0.0):
        """Return ``True`` if all points defined by ``other`` are contained.

        Parameters
        ----------
        other :
            Collection of points to be tested. Can be given as a single
            point, a ``(d, N)`` array-like where ``d`` is the
            number of dimensions, or a length-``d`` `meshgrid` tuple.
        atol : float, optional
            The maximum allowed distance in 'inf'-norm between the
            other set and this interval product.

        Returns
        -------
        contains : bool
            ``True`` if all points are contained, ``False`` otherwise.

        Examples
        --------
        >>> min_pt, max_pt = [-1, 0, 2], [-0.5, 0, 3]
        >>> rbox = IntervalProd(min_pt, max_pt)

        Arrays are expected in ``(ndim, npoints)`` shape:

        >>> arr = np.array([[-1, 0, 2],   # defining one point at a time
        ...                 [-0.5, 0, 2]])
        >>> rbox.contains_all(arr.T)
        True

        Implicit meshgrids defined by coordinate vectors:

        >>> from odl.discr.grid import sparse_meshgrid
        >>> vec1 = (-1, -0.9, -0.7)
        >>> vec2 = (0, 0, 0)
        >>> vec3 = (2.5, 2.75, 3)
        >>> mg = sparse_meshgrid(vec1, vec2, vec3)
        >>> rbox.contains_all(mg)
        True

        Works also with an arbitrary iterable:

        >>> rbox.contains_all([[-1, -0.5], # define points by axis
        ...                    [0, 0],
        ...                    [2, 2]])
        True

        Grids are also accepted as input:

        >>> agrid = odl.uniform_sampling(rbox.min_pt, rbox.max_pt, [3, 1, 3])
        >>> rbox.contains_all(agrid)
        True
        """
        atol = float(atol)

        # First try optimized methods
        if other in self:
            return True
        if hasattr(other, 'meshgrid'):
            return self.contains_all(other.meshgrid, atol=atol)
        elif is_valid_input_meshgrid(other, self.ndim):
            vecs = tuple(vec.squeeze() for vec in other)
            mins = np.fromiter((np.min(vec) for vec in vecs), dtype=float)
            maxs = np.fromiter((np.max(vec) for vec in vecs), dtype=float)
            return (np.all(mins >= self.min_pt - atol) and
                    np.all(maxs <= self.max_pt + atol))

        # Convert to array and check each element
        other = np.asarray(other)
        if is_valid_input_array(other, self.ndim):
            if self.ndim == 1:
                mins = np.min(other)
                maxs = np.max(other)
            else:
                mins = np.min(other, axis=1)
                maxs = np.max(other, axis=1)
            return np.all(mins >= self.min_pt) and np.all(maxs <= self.max_pt)
        else:
            return False

    def measure(self, ndim=None):
        """Return the Lebesgue measure of this interval product.

        Parameters
        ----------
        ndim : int, optional
            Dimension of the measure to apply. ``None`` is interpreted
            as `true_ndim`, which always results in a finite and
            positive result (unless the set is a single point).

        Examples
        --------
        >>> min_pt, max_pt = [-1, 2.5, 0], [-0.5, 10, 0]
        >>> rbox = IntervalProd(min_pt, max_pt)
        >>> rbox.measure()
        3.75
        >>> rbox.measure(ndim=3)
        0.0
        >>> rbox.measure(ndim=3) == rbox.volume
        True
        >>> rbox.measure(ndim=1)
        inf
        >>> rbox.measure() == rbox.squeeze().volume
        True
        """
        if self.true_ndim == 0:
            return 0.0

        if ndim is None:
            return self.measure(ndim=self.true_ndim)
        elif ndim < self.true_ndim:
            return np.inf
        elif ndim > self.true_ndim:
            return 0.0
        else:
            return np.prod((self.max_pt - self.min_pt)[self.inondeg])

    def dist(self, point, exponent=2.0):
        """Return the distance of ``point`` to this set.

        Parameters
        ----------
        point : `array-like` or float
            Point whose distance to calculate. Its length must be equal
            to the set's dimension. Can be a float in the 1d case.
        exponent : non-zero float or ``float('inf')``, optional
            Exponent of the norm used in the distance calculation.

        Returns
        -------
        dist : float
            Distance to the interior of the IntervalProd.
            Points strictly inside have distance ``0.0``, points with
            ``NaN`` have distance ``float('inf')``.

        See Also
        --------
        numpy.linalg.norm : norm used to compute the distance

        Examples
        --------
        >>> min_pt, max_pt = [-1, 0, 2], [-0.5, 0, 3]
        >>> rbox = IntervalProd(min_pt, max_pt)
        >>> rbox.dist([-5, 3, 2])
        5.0
        >>> rbox.dist([-5, 3, 2], exponent=float('inf'))
        4.0
        """
        point = np.atleast_1d(point)
        if len(point) != self.ndim:
            raise ValueError('`point` must have length {}, got {}'
                             ''.format(self.ndim, len(point)))

        if np.any(np.isnan(point)):
            return float('inf')

        i_larger = np.where(point > self.max_pt)
        i_smaller = np.where(point < self.min_pt)

        # Access [0] since np.where returns a tuple.
        if len(i_larger[0]) == 0 and len(i_smaller[0]) == 0:
            return 0.0
        else:
            proj = np.concatenate((point[i_larger], point[i_smaller]))
            border = np.concatenate((self.max_pt[i_larger],
                                     self.min_pt[i_smaller]))
            return np.linalg.norm(proj - border, ord=exponent)

    def collapse(self, indices, values):
        """Partly collapse the interval product to single values.

        Note that no changes are made in-place.

        Parameters
        ----------
        indices : int or sequence of ints
            The indices of the dimensions along which to collapse.
        values : `array-like` or float
            The values to which to collapse. Must have the same
            length as ``indices``. Values must lie within the interval
            boundaries.

        Returns
        -------
        collapsed : `IntervalProd`
            The collapsed set.

        Examples
        --------
        >>> min_pt, max_pt = [-1, 0, 2], [-0.5, 1, 3]
        >>> rbox = IntervalProd(min_pt, max_pt)
        >>> rbox.collapse(1, 0)
        IntervalProd([-1.0, 0.0, 2.0], [-0.5, 0.0, 3.0])
        >>> rbox.collapse([1, 2], [0, 2.5])
        IntervalProd([-1.0, 0.0, 2.5], [-0.5, 0.0, 2.5])
        """
        indices = np.atleast_1d(indices).astype('int64', casting='safe')
        values = np.atleast_1d(values)
        if len(indices) != len(values):
            raise ValueError('lengths of indices {} and values {} do not '
                             'match ({} != {})'
                             ''.format(indices, values,
                                       len(indices), len(values)))

        for axis, index in enumerate(indices):
            if not 0 <= index <= self.ndim:
                raise IndexError('in axis {}: index {} out of range 0 --> {}'
                                 ''.format(axis, index, self.ndim - 1))

        if np.any(values < self.min_pt[indices]):
            raise ValueError('values {} not above the lower interval '
                             'boundaries {}'
                             ''.format(values, self.min_pt[indices]))

        if np.any(values > self.max_pt[indices]):
            raise ValueError('values {} not below the upper interval '
                             'boundaries {}'
                             ''.format(values, self.max_pt[indices]))

        b_new = self.min_pt.copy()
        b_new[indices] = values
        e_new = self.max_pt.copy()
        e_new[indices] = values

        return IntervalProd(b_new, e_new)

    def squeeze(self):
        """Remove the degenerate dimensions.

        Note that no changes are made in-place.

        Returns
        -------
        squeezed : `IntervalProd`
            Squeezed set.

        Examples
        --------
        >>> min_pt, max_pt = [-1, 0, 2], [-0.5, 1, 3]
        >>> rbox = IntervalProd(min_pt, max_pt)
        >>> rbox.collapse(1, 0).squeeze()
        IntervalProd([-1.0, 2.0], [-0.5, 3.0])
        >>> rbox.collapse([1, 2], [0, 2.5]).squeeze()
        IntervalProd(-1.0, -0.5)
        >>> rbox.collapse([0, 1, 2], [-1, 0, 2.5]).squeeze()
        IntervalProd([], [])
        """
        b_new = self.min_pt[self.inondeg]
        e_new = self.max_pt[self.inondeg]
        return IntervalProd(b_new, e_new)

    def insert(self, index, other):
        """Insert ``other`` before ``index``.

        The given interval product (``ndim=m``) is inserted into the
        current one (``ndim=n``) before the given index, resulting in a
        new interval product with ``n+m`` dimensions.

        Note that no changes are made in-place.

        Parameters
        ----------
        index : int
            Index of the dimension before which ``other`` is to
            be inserted. Must fulfill ``-ndim <= index <= ndim``.
            Negative indices count backwards from ``self.ndim``.
        other : `IntervalProd`
            Interval product to be inserted.

        Returns
        -------
        newintvp : `IntervalProd`
            Interval product with ``other`` inserted.

        Examples
        --------
        >>> rbox = IntervalProd([-1, 2], [-0.5, 3])
        >>> rbox2 = IntervalProd([0, 0], [1, 0])
        >>> rbox.insert(1, rbox2)
        IntervalProd([-1.0, 0.0, 0.0, 2.0], [-0.5, 1.0, 0.0, 3.0])
        >>> rbox.insert(-1, rbox2)
        IntervalProd([-1.0, 0.0, 0.0, 2.0], [-0.5, 1.0, 0.0, 3.0])
        """
        index, index_in = int(index), index

        if not -self.ndim <= index <= self.ndim:
            raise IndexError('index {0} outside the valid range -{1} --> {1}'
                             ''.format(index_in, self.ndim))
        if index < 0:
            index += self.ndim

        new_min_pt = np.empty(self.ndim + other.ndim)
        new_max_pt = np.empty(self.ndim + other.ndim)

        new_min_pt[: index] = self.min_pt[: index]
        new_max_pt[: index] = self.max_pt[: index]
        new_min_pt[index: index + other.ndim] = other.min_pt
        new_max_pt[index: index + other.ndim] = other.max_pt
        if index < self.ndim:  # Avoid IndexError
            new_min_pt[index + other.ndim:] = self.min_pt[index:]
            new_max_pt[index + other.ndim:] = self.max_pt[index:]

        return IntervalProd(new_min_pt, new_max_pt)

    def append(self, other):
        """Insert at the end.

        Parameters
        ----------
        other : `IntervalProd`
            Set to be appended.

        Returns
        -------
        newintvp : `IntervalProd`
            Interval product with ``other`` appended.

        Examples
        --------
        >>> rbox = IntervalProd([-1, 2], [-0.5, 3])
        >>> rbox.append(IntervalProd(-1.0, 0.0))
        IntervalProd([-1.0, 2.0, -1.0], [-0.5, 3.0, 0.0])

        See Also
        --------
        insert
        """
        return self.insert(self.ndim, other)

    def corners(self, order='C'):
        """Return the corner points as a single array.

        Parameters
        ----------
        order : {'C', 'F'}, optional
            Ordering of the axes in which the corners appear in
            the output. ``'C'`` means that the first axis varies slowest
            and the last one fastest, vice versa in ``'F'`` ordering.

        Returns
        -------
        corners : `numpy.ndarray`
            Array containing the corner coordinates. The size of the
            array is ``2^m x ndim``, where ``m`` is the number of
            non-degenerate axes, i.e. the corners are stored as rows.

        Examples
        --------
        >>> rbox = IntervalProd([-1, 2, 0], [-0.5, 3, 0.5])
        >>> rbox.corners()
        array([[-1. ,  2. ,  0. ],
               [-1. ,  2. ,  0.5],
               [-1. ,  3. ,  0. ],
               [-1. ,  3. ,  0.5],
               [-0.5,  2. ,  0. ],
               [-0.5,  2. ,  0.5],
               [-0.5,  3. ,  0. ],
               [-0.5,  3. ,  0.5]])
        >>> rbox.corners(order='F')
        array([[-1. ,  2. ,  0. ],
               [-0.5,  2. ,  0. ],
               [-1. ,  3. ,  0. ],
               [-0.5,  3. ,  0. ],
               [-1. ,  2. ,  0.5],
               [-0.5,  2. ,  0.5],
               [-1. ,  3. ,  0.5],
               [-0.5,  3. ,  0.5]])
        """
        from odl.discr.grid import TensorGrid

        minmax_vecs = [0] * self.ndim
        for axis in self.ideg:
            minmax_vecs[axis] = self.min_pt[axis]
        for axis in self.inondeg:
            minmax_vecs[axis] = (self.min_pt[axis], self.max_pt[axis])

        minmax_grid = TensorGrid(*minmax_vecs)
        return minmax_grid.points(order=order)

    def __len__(self):
        """Return ``len(self)``."""
        return self.ndim

    def __getitem__(self, indices):
        """Return ``self[indices]``

        Parameters
        ----------
        indices : index expression
            Object determining which parts of the interval product
            to extract.

        Returns
        -------
        subinterval : `IntervalProd`
            Interval product corresponding to the given indices.

        Examples
        --------
        >>> rbox = IntervalProd([-1, 2, 0], [-0.5, 3, 0.5])

        Indexing by integer selects single axes:

        >>> rbox[0]
        IntervalProd(-1.0, -0.5)

        With slices, multiple axes can be selected:

        >>> rbox[:]
        IntervalProd([-1.0, 2.0, 0.0], [-0.5, 3.0, 0.5])
        >>> rbox[::2]
        IntervalProd([-1.0, 0.0], [-0.5, 0.5])

        A list of integers can be used for free combinations of axes:

        >>> rbox[[0, 1, 0]]
        IntervalProd([-1.0, 2.0, -1.0], [-0.5, 3.0, -0.5])
        """
        return IntervalProd(self.min_pt[indices], self.max_pt[indices])

    def __pos__(self):
        """Return ``+self``."""
        return self

    def __neg__(self):
        """Return ``-self``."""
        return type(self)(-self.max_pt, -self.min_pt)

    def __add__(self, other):
        """Return ``self + other``."""
        if isinstance(other, IntervalProd):
            if self.ndim != other.ndim:
                raise ValueError('addition not possible for {} and {}: '
                                 'dimension mismatch ({} != {})'
                                 ''.format(self, other, self.ndim, other.ndim))
            return type(self)(self.min_pt + other.min_pt,
                              self.max_pt + other.max_pt)
        elif np.isscalar(other):
            return type(self)(self.min_pt + other, self.max_pt + other)
        else:
            return NotImplemented

    def __sub__(self, other):
        """Return ``self - other``."""
        return self + (-other)

    def __mul__(self, other):
        """Return ``self * other``."""
        if isinstance(other, IntervalProd):
            if self.ndim != other.ndim:
                raise ValueError('multiplication not possible for {} and'
                                 '{}: dimension mismatch ({} != {})'
                                 ''.format(self, other, self.ndim, other.ndim))

            comp_mat = np.empty([self.ndim, 4])
            comp_mat[:, 0] = self.min_pt * other.min_pt
            comp_mat[:, 1] = self.min_pt * other.max_pt
            comp_mat[:, 2] = self.max_pt * other.min_pt
            comp_mat[:, 3] = self.max_pt * other.max_pt
            new_min_pt = np.min(comp_mat, axis=1)
            new_max_pt = np.max(comp_mat, axis=1)
            return type(self)(new_min_pt, new_max_pt)
        elif np.isscalar(other):
            vec1 = self.min_pt * other
            vec2 = self.max_pt * other
            return type(self)(np.minimum(vec1, vec2), np.maximum(vec1, vec2))
        else:
            return NotImplemented

    def __div__(self, other):
        """Return ``self / other``."""
        return self * (1.0 / other)

    __truediv__ = __div__

    def __rdiv__(self, other):
        """Return ``other / self``."""
        if np.isscalar(other):
            for axis, (xmin, xmax) in enumerate(zip(self.min_pt, self.max_pt)):
                if xmin <= 0 and xmax >= 0:
                    raise ValueError('in axis {}: interval {} contains 0, '
                                     'division not possible'
                                     .format(axis, [xmin, xmax]))

            vec1 = other / self.min_pt
            vec2 = other / self.max_pt
            return type(self)(np.minimum(vec1, vec2), np.maximum(vec1, vec2))
        else:
            return NotImplemented

    __rtruediv__ = __rdiv__

    def __repr__(self):
        """Return ``repr(self)``."""
        if self.ndim == 1:
            return '{}({!r}, {!r})'.format(self.__class__.__name__,
                                           self.min_pt[0],
                                           self.max_pt[0])
        else:
            return '{}({}, {})'.format(self.__class__.__name__,
                                       array1d_repr(self.min_pt),
                                       array1d_repr(self.max_pt))

    def __str__(self):
        """Return ``str(self)``."""
        return ' x '.join('[{}, {}]'.format(xmin, xmax)
                          for xmin, xmax in zip(self.min_pt, self.max_pt))


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Standardized tests for `Operator`'s."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import object

import numpy as np

from odl.diagnostics.examples import samples
from odl.operator import power_method_opnorm
from odl.util.testutils import FailCounter


__all__ = ('OperatorTest',)


class OperatorTest(object):

    """Automated tests for `Operator` implementations.

    This class allows users to automatically test various
    features of an Operator such as linearity, the adjoint definition and
    definition of the derivative.
    """

    def __init__(self, operator, operator_norm=None, verbose=True, tol=1e-5):
        """Initialize a new instance.

        Parameters
        ----------
        operator : `Operator`
            The operator to run tests on
        operator_norm : float, optional
            The norm of the operator, used for error estimates. If
            ``None`` is given, the norm is estimated during
            initialization.
        verbose : bool, optional
            If ``True``, print additional info text.
        tol : float, optional
            Tolerance parameter used as a base for the actual tolerance
            in the tests. Depending on the expected accuracy, the actual
            tolerance used in a test can be a factor times this number.
        """
        self.operator = operator
        self.verbose = False
        if operator_norm is None:
            self.operator_norm = self.norm()
        else:
            self.operator_norm = float(operator_norm)

        self.verbose = bool(verbose)
        self.tol = float(tol)

    def log(self, message):
        """Print message if ``self.verbose == True``."""
        if self.verbose:
            print(message)

    def norm(self):
        """Estimate the operator norm of the operator.

        The norm is estimated by calculating

        ``A(x).norm() / x.norm()``

        for some nonzero ``x``

        Returns
        -------
        norm : float
            Estimate of operator norm

        References
        ----------
        Wikipedia article on `Operator norm
        <https://en.wikipedia.org/wiki/Operator_norm>`_.
        """
        self.log('\n== Calculating operator norm ==\n')

        operator_norm = max(power_method_opnorm(self.operator, maxiter=2,
                                                xstart=x)
                            for name, x in samples(self.operator.domain)
                            if name != 'Zero')

        self.log('Norm is at least: {}'.format(operator_norm))
        self.operator_norm = operator_norm
        return operator_norm

    def self_adjoint(self):
        """Verify ``<Ax, y> == <x, Ay>``."""
        left_inner_vals = []
        right_inner_vals = []

        with FailCounter(
                test_name='Verifying the identity <Ax, y> = <x, Ay>',
                err_msg='error = |<Ax, y> - <x, Ay>| / ||A|| ||x|| ||y||',
                logger=self.log) as counter:

            for [name_x, x], [name_y, y] in samples(self.operator.domain,
                                                    self.operator.range):
                x_norm = x.norm()
                y_norm = y.norm()

                l_inner = self.operator(x).inner(y)
                r_inner = x.inner(self.operator(y))

                denom = self.operator_norm * x_norm * y_norm
                error = 0 if denom == 0 else abs(l_inner - r_inner) / denom

                if error > self.tol:
                    counter.fail('x={:25s} y={:25s} : error={:6.5f}'
                                 ''.format(name_x, name_y, error))

                left_inner_vals.append(l_inner)
                right_inner_vals.append(r_inner)

        scale = np.polyfit(left_inner_vals, right_inner_vals, 1)[0]
        self.log('\nThe adjoint seems to be scaled according to:')
        self.log('(x, Ay) / (Ax, y) = {}. Should be 1.0'.format(scale))

    def _adjoint_definition(self):
        """Verify ``<Ax, y> == <x, A^* y>``."""
        left_inner_vals = []
        right_inner_vals = []

        with FailCounter(
                test_name='Verifying the identity <Ax, y> = <x, A^T y>',
                err_msg='error = |<Ax, y< - <x, A^* y>| / ||A|| ||x|| ||y||',
                logger=self.log) as counter:

            for [name_x, x], [name_y, y] in samples(self.operator.domain,
                                                    self.operator.range):
                x_norm = x.norm()
                y_norm = y.norm()

                l_inner = self.operator(x).inner(y)
                r_inner = x.inner(self.operator.adjoint(y))

                denom = self.operator_norm * x_norm * y_norm
                error = 0 if denom == 0 else abs(l_inner - r_inner) / denom

                if error > self.tol:
                    counter.fail('x={:25s} y={:25s} : error={:6.5f}'
                                 ''.format(name_x, name_y, error))

                left_inner_vals.append(l_inner)
                right_inner_vals.append(r_inner)

        scale = np.polyfit(left_inner_vals, right_inner_vals, 1)[0]
        self.log('\nThe adjoint seems to be scaled according to:')
        self.log('(x, A^T y) / (Ax, y) = {}. Should be 1.0'.format(scale))

    def _adjoint_of_adjoint(self):
        """Verify ``(A^*)^* == A``"""
        try:
            self.operator.adjoint.adjoint
        except AttributeError:
            print('A^* has no adjoint')
            return

        if self.operator.adjoint.adjoint is self.operator:
            self.log('(A^*)^* == A')
            return

        with FailCounter(
                test_name='\nVerifying the identity Ax = (A^*)^* x',
                err_msg='error = ||Ax - (A^*)^* x|| / ||A|| ||x||',
                logger=self.log) as counter:
            for [name_x, x] in self.operator.domain.examples:
                opx = self.operator(x)
                op_adj_adj_x = self.operator.adjoint.adjoint(x)

                denom = self.operator_norm * x.norm()
                if denom == 0:
                    error = 0
                else:
                    error = (opx - op_adj_adj_x).norm() / denom

                if error > self.tol:
                    counter.fail('x={:25s} : error={:6.5f}'
                                 ''.format(name_x, error))

    def adjoint(self):
        """Verify that `Operator.adjoint` works appropriately.

        References
        ----------
        Wikipedia article on `Adjoint
        <https://en.wikipedia.org/wiki/Adjoint>`_.
        """
        try:
            self.operator.adjoint
        except NotImplementedError:
            print('Operator has no adjoint')
            return

        self.log('\n== Verifying operator adjoint ==\n')

        domain_range_ok = True
        if self.operator.domain != self.operator.adjoint.range:
            print('*** ERROR: A.domain != A.adjoint.range ***')
            domain_range_ok = False

        if self.operator.range != self.operator.adjoint.domain:
            print('*** ERROR: A.range != A.adjoint.domain ***')
            domain_range_ok = False

        if domain_range_ok:
            self.log('Domain and range of adjoint are OK.')
        else:
            print('Domain and range of adjoint are not OK, exiting.')
            return

        self._adjoint_definition()
        self._adjoint_of_adjoint()

    def _derivative_convergence(self):
        """Verify that the derivative is a first-order approximation.

        The code verifies if

            ``||A(x+c*p) - A(x) - A'(x)(c*p)|| / c = o(c)``

        for ``c --> 0``.
        """
        with FailCounter(
                test_name='Verifying that derivative is a first-order '
                          'approximation',
                err_msg="error = inf_c ||A(x+c*p)-A(x)-A'(x)(c*p)|| / c",
                logger=self.log) as counter:
            for [name_x, x], [name_dx, dx] in samples(self.operator.domain,
                                                      self.operator.domain):
                # Precompute some values
                deriv = self.operator.derivative(x)
                derivdx = deriv(dx)
                opx = self.operator(x)

                c = 1e-4  # initial step
                derivative_ok = False

                minerror = float('inf')
                while c > 1e-14:
                    exact_step = self.operator(x + dx * c) - opx
                    expected_step = c * derivdx
                    err = (exact_step - expected_step).norm() / c

                    # Need to be slightly more generous here due to possible
                    # numerical instabilities.
                    # TODO: perform more tests to find a good threshold here.
                    if err < 10 * self.tol:
                        derivative_ok = True
                        break
                    else:
                        minerror = min(minerror, err)

                    c /= 10.0

                if not derivative_ok:
                    counter.fail('x={:15s} p={:15s}, error={}'
                                 ''.format(name_x, name_dx, minerror))

    def derivative(self):
        """Verify that `Operator.derivative` works appropriately.

        The code verifies if

            ``||A(x+c*p) - A(x) - A'(x)(c*p)|| / c = o(c)``

        for ``c --> 0`` using a selection of elements ``x`` and ``p``.

        References
        ----------
        Wikipedia article on `Derivative
        <https://en.wikipedia.org/wiki/Derivative>`_.
        Wikipedia article on `Frechet derivative
        <https://en.wikipedia.org/wiki/Fr%C3%A9chet_derivative>`_.
        """
        self.log('\n== Verifying operator derivative  ==')

        try:
            deriv = self.operator.derivative(self.operator.domain.zero())

            if not deriv.is_linear:
                print('Derivative is not a linear operator')
                return
        except NotImplementedError:
            print('Operator has no derivative')
            return

        if self.operator.is_linear and deriv is self.operator:
            self.log('A is linear and A.derivative is A')
            return

        self._derivative_convergence()

    def _scale_invariance(self):
        """Verify ``A(c*x) = c * A(x)``."""
        with FailCounter(
                test_name='Verifying homogeneity under scalar multiplication',
                err_msg='error = ||A(c*x)-c*A(x)|| / |c| ||A|| ||x||',
                logger=self.log) as counter:
            for [name_x, x], [_, scale] in samples(self.operator.domain,
                                                   self.operator.domain.field):
                opx = self.operator(x)
                scaled_opx = self.operator(scale * x)

                denom = self.operator_norm * scale * x.norm()
                error = (0 if denom == 0
                         else (scaled_opx - opx * scale).norm() / denom)

                if error > self.tol:
                    counter.fail('x={:25s} scale={:7.2f} error={:6.5f}'
                                 ''.format(name_x, scale, error))

    def _addition_invariance(self):
        """Verify ``A(x+y) = A(x) + A(y)``."""
        with FailCounter(
                test_name='Verifying distributivity under vector addition',
                err_msg='error = ||A(x+y) - A(x) - A(y)|| / '
                        '||A||(||x|| + ||y||)',
                logger=self.log) as counter:
            for [name_x, x], [name_y, y] in samples(self.operator.domain,
                                                    self.operator.domain):
                opx = self.operator(x)
                opy = self.operator(y)
                opxy = self.operator(x + y)

                denom = self.operator_norm * (x.norm() + y.norm())
                error = (0 if denom == 0
                         else (opxy - opx - opy).norm() / denom)

                if error > self.tol:
                    counter.fail('x={:25s} y={:25s} error={:6.5f}'
                                 ''.format(name_x, name_y, error))

    def linear(self):
        """Verify that the operator is actually linear."""
        if not self.operator.is_linear:
            print('Operator is not linear')
            return

        self.log('\n== Verifying operator linearity ==\n')

        # Test if zero gives zero
        result = self.operator(self.operator.domain.zero())
        result_norm = result.norm()
        if result_norm != 0.0:
            print("||A(0)||={:6.5f}. Should be 0.0000".format(result_norm))

        self._scale_invariance()
        self._addition_invariance()

    def run_tests(self):
        """Run all tests on this operator."""
        print('\n== RUNNING ALL TESTS ==')
        print('Operator = {}'.format(self.operator))

        self.norm()

        if self.operator.is_linear:
            self.linear()
            self.adjoint()
        else:
            self.derivative()

    def __str__(self):
        return '{}({})'.format(self.__class__.__name__, self.operator)

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self.operator)

if __name__ == '__main__':
    # pylint: disable=wrong-import-position

    import odl
    space = odl.uniform_discr([0, 0], [1, 1], [3, 3])
    # Linear operator
    I = odl.IdentityOperator(space)
    OperatorTest(I, verbose=False).run_tests()

    # Nonlinear operator op(x) = x**4
    op = odl.PowerOperator(space, 4)
    OperatorTest(op).run_tests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import int

import numpy as np


def euler_matrix(*angles):
    """Rotation matrix in 2 and 3 dimensions.

    Compute the Euler rotation matrix from angles given in radians.
    Its rows represent the canonical unit vectors as seen from the
    rotated system while the columns are the rotated unit vectors as
    seen from the canonical system.

    Parameters
    ----------
    angle1,...,angleN : float
        One angle results in a (2x2) matrix representing a
        counter-clockwise rotation. Two or three angles result in a
        (3x3) matrix and are interpreted as Euler angles of a 3d
        rotation according to the 'ZXZ' rotation order, see the
        Wikipedia article `Euler angles`_.

    Returns
    -------
    mat : `numpy.ndarray`, shape ``(2, 2)`` or ``(3, 3)``
        The rotation matrix

    .. _Euler angles:
        https://en.wikipedia.org/wiki/Euler_angles#Rotation_matrix
    """
    if len(angles) == 1:
        phi = float(angles[0])
        theta = psi = 0.
        ndim = 2
    elif len(angles) == 2:
        phi = float(angles[0])
        theta = float(angles[1])
        psi = 0.
        ndim = 3
    elif len(angles) == 3:
        phi = float(angles[0])
        theta = float(angles[1])
        psi = float(angles[2])
        ndim = 3
    else:
        raise ValueError('number of angles must be between 1 and 3')

    cph = np.cos(phi)
    sph = np.sin(phi)
    cth = np.cos(theta)
    sth = np.sin(theta)
    cps = np.cos(psi)
    sps = np.sin(psi)

    if ndim == 2:
        mat = np.array([[cph, -sph],
                        [sph, cph]])
    else:
        mat = np.array([
            [cph * cps - sph * cth * sps,
             -cph * sps - sph * cth * cps,
             sph * sth],
            [sph * cps + cph * cth * sps,
             -sph * sps + cph * cth * cps,
             -cph * sth],
            [sth * sps,
             sth * cps,
             cth]])

    return mat


def axis_rotation(axis, angle, vectors):
    """Rotate a vector or an array of vectors around an axis in 3d.

    The rotation is computed by `Rodriguez' rotation formula`_.

    Parameters
    ----------
    axis : `array-like`, shape ``(3,)``
        The rotation axis, assumed to be a unit vector
    angle : float
        The rotation angle
    vectors : `array-like`, shape ``(3,)`` or ``(N, 3)``
        The vector(s) to be rotated

    Returns
    -------
    rot_vec : `numpy.ndarray`
        The rotated vector(s)

    .. _Rodriguez' rotation formula:
        https://en.wikipedia.org/wiki/Rodrigues'_rotation_formula
    """
    if not (hasattr(vectors, 'shape') and hasattr(vectors, 'ndim')):
        vectors = np.asarray(vectors)

    if not (vectors.shape == (3,) or (vectors.ndim == 2 and
                                      vectors.shape[1] == 3)):
        raise ValueError('`vector` shape {} not (3,) or (N, 3)'
                         ''.format(vectors.shape))

    if not hasattr(axis, 'shape'):
        axis = np.asarray(axis)

    if axis.shape != (3,):
        raise ValueError('`axis` shape {} not (3,)'.format(axis.shape))

    angle = float(angle)

    if np.isclose(angle / (2 * np.pi), int(angle / (2 * np.pi)), atol=1e-15):
        return vectors.copy()
    else:
        cos_ang = np.cos(angle)
        sin_ang = np.sin(angle)
        scal = np.asarray(np.dot(vectors, axis))
        cross = np.asarray(np.cross(vectors, axis))
        cross *= sin_ang

        rot_vecs = cos_ang * vectors
        rot_vecs += (1. - cos_ang) * scal[:, None] * axis[None, :]
        rot_vecs += cross


def axis_rotation_matrix(axis, angle):
    """Matrix of the rotation around an axis in 3d.

    The matrix is computed according to `Rodriguez' rotation formula`_.

    Parameters
    ----------
    axis : `array-like`, shape ``(3,)``
        The rotation axis, assumed to be a unit vector
    angle : float
        The rotation angle

    Returns
    -------
    mat : `numpy.ndarray`, shape ``(3, 3)``
        The axis rotation matrix

    .. _Rodriguez' rotation formula:
        https://en.wikipedia.org/wiki/Rodrigues'_rotation_formula
    """
    if not hasattr(axis, 'shape'):
        axis = np.asarray(axis)

    if axis.shape != (3,):
        raise ValueError('`axis` shape {} not (3,)'.format(axis.shape))

    angle = float(angle)

    cross_mat = np.matrix([[0, -axis[2], axis[1]],
                           [axis[2], 0, -axis[0]],
                           [-axis[1], axis[0], 0]])
    dy_mat = np.asmatrix(np.outer(axis, axis))
    id_mat = np.asmatrix(np.eye(3))
    cos_ang = np.cos(angle)
    sin_ang = np.sin(angle)

    return cos_ang * id_mat + (1. - cos_ang) * dy_mat + sin_ang * cross_mat


def is_rotation_matrix(mat, show_diff=False):
    from scipy.linalg import det, norm

    dim = mat.shape[0]
    if dim != mat.shape[1]:
        return False

    determ = det(mat)
    right_handed = (np.abs(determ - 1.) < 1E-10)
    orthonorm_diff = mat * mat.T - np.eye(dim)
    diff_norm = norm(orthonorm_diff, 2)
    orthonormal = (diff_norm < 1E-10)
    if not right_handed or not orthonormal:
        if show_diff:
            print('matrix S:\n', mat)
            print('det(S): ', determ)
            print('S*S.T - eye:\n', orthonorm_diff)
            print('2-norm of difference: ', diff_norm)
        return False
    return True


def angles_from_matrix(rot_matrix):
    if rot_matrix.shape == (2, 2):
        theta = np.atan2(rot_matrix[1, 0], rot_matrix[0, 0])
        return theta,
    elif rot_matrix.shape == (3, 3):
        if rot_matrix[2, 2] == 1.:  # cannot use last row and column
            theta = 0.
            # upper-left block is 2d rotation for phi + psi, so one needs
            # to be fixed
            psi = 0.
            phi = np.atan2(rot_matrix[1, 0], rot_matrix[0, 0])
            if phi < 0:
                phi += 2 * np.pi  # in [0, 2pi)
        else:
            phi = np.atan2(rot_matrix[0, 2], -rot_matrix[1, 2])
            psi = np.atan2(rot_matrix[2, 0], rot_matrix[2, 1])
            theta = np.acos(rot_matrix[2, 2])

            if phi < 0. or psi < 0.:
                phi += np.pi
                psi += np.pi
                theta = -theta

        return phi, theta, psi

    else:
        raise ValueError('shape of `rot_matrix` is {}, expected (2, 2) '
                         'or (3, 3)'.format(rot_matrix.shape))


def to_lab_sys(vec_in_local_coords, local_sys):
    vec_in_local_coords = np.array(vec_in_local_coords)
    trafo_matrix = np.matrix(local_sys).T
    return np.dot(trafo_matrix, vec_in_local_coords)


def to_local_sys(vec_in_lab_coords, local_sys):
    vec_in_lab_coords = np.array(vec_in_lab_coords)
    trafo_matrix = np.matrix(local_sys)
    return np.dot(trafo_matrix, vec_in_lab_coords)


def perpendicular_vector(vec):
    """Return a vector perpendicular to ``vec``.

    Parameters
    ----------
    vec : `array-like`
        Vector of arbitrary length.

    Returns
    -------
    perp_vec : `numpy.ndarray`
        Array of same size such that ``<vec, perp_vec> == 0``

    Examples
    --------
    Works in 2d:

    >>> perpendicular_vector([1, 0])
    array([ 0.,  1.])
    >>> perpendicular_vector([0, 1])
    array([-1.,  0.])

    And in 3d:

    >>> perpendicular_vector([1, 0, 0])
    array([ 0.,  1.,  0.])
    >>> perpendicular_vector([0, 1, 0])
    array([-1.,  0.,  0.])
    >>> perpendicular_vector([0, 0, 1])
    array([ 1.,  0.,  0.])
    """
    vec = np.asarray(vec)

    if np.all(vec == 0):
        raise ValueError('zero vector')

    result = np.zeros(vec.shape)
    if np.any(vec[:2] != 0):
        result[:2] = [-vec[1], vec[0]]
    else:
        result[0] = 1

    return result / np.linalg.norm(result)


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Discrete wavelet transformation on L2 spaces."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import str, super

import numpy as np

from odl.discr import DiscreteLp
from odl.operator import Operator
from odl.trafos.backends.pywt_bindings import (
    PYWT_AVAILABLE, PAD_MODES_ODL2PYWT,
    pywt_pad_mode, pywt_wavelet, pywt_flat_coeff_size, pywt_coeff_shapes,
    pywt_flat_array_from_coeffs, pywt_coeffs_from_flat_array,
    pywt_multi_level_decomp, pywt_multi_level_recon)

__all__ = ('WaveletTransform', 'WaveletTransformInverse')


_SUPPORTED_WAVELET_IMPLS = ()
if PYWT_AVAILABLE:
    _SUPPORTED_WAVELET_IMPLS += ('pywt',)


class WaveletTransformBase(Operator):

    """Base class for discrete wavelet transforms.

    This abstract class is intended to share code between the forward,
    inverse and adjoint wavelet transforms.
    """

    def __init__(self, space, wavelet, nlevels, variant, pad_mode='constant',
                 pad_const=0, impl='pywt'):
        """Initialize a new instance.

        Parameters
        ----------
        space : `DiscreteLp`
            Domain of the forward wavelet transform (the "image domain").
            In the case of ``variant in ('inverse', 'adjoint')``, this
            space is the range of the operator.
        wavelet : string or `pywt.Wavelet`
            Specification of the wavelet to be used in the transform.
            If a string is given, it is converted to a `pywt.Wavelet`.
            Use `pywt.wavelist` to get a list of available wavelets.

            Possible wavelet families are:

            ``'haar'``: Haar

            ``'db'``: Daubechies

            ``'sym'``: Symlets

            ``'coif'``: Coiflets

            ``'bior'``: Biorthogonal

            ``'rbio'``: Reverse biorthogonal

            ``'dmey'``: Discrete FIR approximation of the Meyer wavelet

        nlevels : positive int
            Number of scaling levels to be used in the decomposition. The
            maximum number of levels can be calculated with
            `pywt.dwt_max_level`.
        variant : {'forward', 'inverse', 'adjoint'}
            Wavelet transform variant to be created.
        pad_mode : string, optional
            Method to be used to extend the signal.

            ``'constant'``: Fill with ``pad_const``.

            ``'symmetric'``: Reflect at the boundaries, not doubling the
            outmost values.

            ``'periodic'``: Fill in values from the other side, keeping
            the order.

            ``'order0'``: Extend constantly with the outmost values
            (ensures continuity).

            ``'order1'``: Extend with constant slope (ensures continuity of
            the first derivative). This requires at least 2 values along
            each axis where padding is applied.

            ``'pywt_per'``:  like ``'periodic'``-padding but gives the smallest
            possible number of decomposition coefficients.
            Only available with ``impl='pywt'``, See `pywt.MODES.modes`.

        pad_const : float, optional
            Constant value to use if ``pad_mode == 'constant'``. Ignored
            otherwise. Constants other than 0 are not supported by the
            ``pywt`` back-end.
        impl : {'pywt'}, optional
            Back-end for the wavelet transform.
        """
        if not isinstance(space, DiscreteLp):
            raise TypeError('`space` {!r} is not a `DiscreteLp` instance.'
                            ''.format(space))

        self.__nlevels, nlevels_in = int(nlevels), nlevels
        if self.nlevels != nlevels_in:
            raise ValueError('`nlevels` must be integer, got {}'
                             ''.format(nlevels_in))

        self.__impl, impl_in = str(impl).lower(), impl
        if self.impl not in _SUPPORTED_WAVELET_IMPLS:
            raise ValueError("`impl` '{}' not supported".format(impl_in))

        self.__wavelet = getattr(wavelet, 'name', str(wavelet).lower())
        self.__pad_mode = str(pad_mode).lower()
        self.__pad_const = space.field.element(pad_const)

        if self.impl == 'pywt':
            self.pywt_pad_mode = pywt_pad_mode(pad_mode, pad_const)
            self.pywt_wavelet = pywt_wavelet(self.wavelet)
            coeff_size = pywt_flat_coeff_size(space.shape, wavelet,
                                              self.nlevels, self.pywt_pad_mode)
            coeff_space = space.dspace_type(coeff_size, dtype=space.dtype)
        else:
            raise RuntimeError("bad `impl` '{}'".format(self.impl))

        variant, variant_in = str(variant).lower(), variant
        if variant not in ('forward', 'inverse', 'adjoint'):
            raise ValueError("`variant` '{}' not understood"
                             "".format(variant_in))

        if variant == 'forward':
            super().__init__(domain=space, range=coeff_space, linear=True)
        else:
            super().__init__(domain=coeff_space, range=space, linear=True)

    @property
    def impl(self):
        """Implementation back-end of this wavelet transform."""
        return self.__impl

    @property
    def nlevels(self):
        """Number of scaling levels in this wavelet transform."""
        return self.__nlevels

    @property
    def wavelet(self):
        """Name of the wavelet used in this wavelet transform."""
        return self.__wavelet

    @property
    def pad_mode(self):
        """Padding mode used for extending input beyond its boundary."""
        return self.__pad_mode

    @property
    def pad_const(self):
        """Value for extension used in ``'constant'`` padding mode."""
        return self.__pad_const

    @property
    def is_orthogonal(self):
        """Whether or not the wavelet basis is orthogonal."""
        return self.pywt_wavelet.orthogonal

    @property
    def is_biorthogonal(self):
        """Whether or not the wavelet basis is bi-orthogonal."""
        return self.pywt_wavelet.biorthogonal


class WaveletTransform(WaveletTransformBase):

    """Discrete wavelet transform between discretized Lp spaces."""

    def __init__(self, domain, wavelet, nlevels, pad_mode='constant',
                 pad_const=0, impl='pywt'):
        """Initialize a new instance.

        Parameters
        ----------
        domain : `DiscreteLp`
            Domain of the wavelet transform (the "image domain").
        wavelet : string or `pywt.Wavelet`
            Specification of the wavelet to be used in the transform.
            If a string is given, it is converted to a `pywt.Wavelet`.
            Use `pywt.wavelist` to get a list of available wavelets.

            Possible wavelet families are:

            ``'haar'``: Haar

            ``'db'``: Daubechies

            ``'sym'``: Symlets

            ``'coif'``: Coiflets

            ``'bior'``: Biorthogonal

            ``'rbio'``: Reverse biorthogonal

            ``'dmey'``: Discrete FIR approximation of the Meyer wavelet

        nlevels : positive int
            Number of scaling levels to be used in the decomposition. The
            maximum number of levels can be calculated with
            `pywt.dwt_max_level`.
        pad_mode : string, optional
            Method to be used to extend the signal.

            ``'constant'``: Fill with ``pad_const``.

            ``'symmetric'``: Reflect at the boundaries, not doubling the
            outmost values.

            ``'periodic'``: Fill in values from the other side, keeping
            the order.

            ``'order0'``: Extend constantly with the outmost values
            (ensures continuity).

            ``'order1'``: Extend with constant slope (ensures continuity of
            the first derivative). This requires at least 2 values along
            each axis where padding is applied.

            ``'pywt_per'``:  like ``'periodic'`` padding, but gives the
            smallest possible number of decomposition coefficients.
            Only available with ``impl='pywt'``, See `pywt.Modes.modes`.

        pad_const : float, optional
            Constant value to use if ``pad_mode == 'constant'``. Ignored
            otherwise. Constants other than 0 are not supported by the
            ``pywt`` back-end.
        impl : {'pywt'}, optional
            Backend for the wavelet transform.

        Examples
        --------
        Compute a very simple wavelet transform in a discrete 2D space with
        4 sampling points per axis:

        >>> space = odl.uniform_discr([0, 0], [1, 1], (4, 4))
        >>> wavelet_trafo = odl.trafos.WaveletTransform(
        ...     domain=space, nlevels=1, wavelet='haar')
        >>> wavelet_trafo.is_biorthogonal
        True
        >>> decomp = wavelet_trafo([[1, 1, 1, 1],
        ...                         [0, 0, 0, 0],
        ...                         [0, 0, 1, 1],
        ...                         [1, 0, 1, 0]])
        >>> print(decomp)
        [1.0, 1.0, 0.5, ..., 0.0, -0.5, -0.5]
        >>> decomp.shape
        (16,)
        """
        super().__init__(space=domain, wavelet=wavelet, nlevels=nlevels,
                         variant='forward', pad_mode=pad_mode,
                         pad_const=pad_const, impl=impl)

    def _call(self, x):
        """Return wavelet transform of ``x``."""
        if self.impl == 'pywt':
            coeff_list = pywt_multi_level_decomp(
                x, wavelet=self.pywt_wavelet, nlevels=self.nlevels,
                mode=self.pywt_pad_mode)
            return pywt_flat_array_from_coeffs(coeff_list)
        else:
            raise RuntimeError("bad `impl` '{}'".format(self.impl))

    @property
    def adjoint(self):
        """Adjoint wavelet transform.

        Returns
        -------
        adjoint : `WaveletTransformInverse`
            If the transform is orthogonal, the adjoint is the inverse.

        Raises
        ------
        OpNotImplementedError
            if `is_orthogonal` is ``False``
        """
        if self.is_orthogonal:
            return self.inverse
        else:
            # TODO: put adjoint here
            return super().adjoint

    @property
    def inverse(self):
        """Inverse wavelet transform.

        Returns
        -------
        inverse : `WaveletTransformInverse`

        See Also
        --------
        adjoint
        """
        return WaveletTransformInverse(
            range=self.domain, nlevels=self.nlevels, wavelet=self.pywt_wavelet,
            pad_mode=self.pad_mode, pad_const=self.pad_const, impl=self.impl)


class WaveletTransformInverse(WaveletTransformBase):

    """Discrete inverse wavelet trafo between discrete L2 spaces.

    See Also
    --------
    WaveletTransform
    """

    def __init__(self, range, nlevels, wavelet, pad_mode='constant',
                 pad_const=0, impl='pywt'):
        """Initialize a new instance.

         Parameters
        ----------
        range : `DiscreteLp`
            Domain of the forward wavelet transform (the "image domain"),
            which is the range of this inverse transform.
        wavelet : string or `pywt.Wavelet`
            Specification of the wavelet to be used in the transform.
            If a string is given, it is converted to a `pywt.Wavelet`.
            Use `pywt.wavelist` to get a list of available wavelets.

            Possible wavelet families are:

            ``'haar'``: Haar

            ``'db'``: Daubechies

            ``'sym'``: Symlets

            ``'coif'``: Coiflets

            ``'bior'``: Biorthogonal

            ``'rbio'``: Reverse biorthogonal

            ``'dmey'``: Discrete FIR approximation of the Meyer wavelet

        nlevels : positive int
            Number of scaling levels to be used in the decomposition. The
            maximum number of levels can be calculated with
            `pywt.dwt_max_level`.
        variant : {'forward', 'inverse', 'adjoint'}
            Wavelet transform variant to be created.
        pad_mode : string, optional
            Method to be used to extend the signal.

            ``'constant'``: Fill with ``pad_const``.

            ``'symmetric'``: Reflect at the boundaries, not doubling the
            outmost values.

            ``'periodic'``: Fill in values from the other side, keeping
            the order.

            ``'order0'``: Extend constantly with the outmost values
            (ensures continuity).

            ``'order1'``: Extend with constant slope (ensures continuity of
            the first derivative). This requires at least 2 values along
            each axis where padding is applied.

            ``'pywt_per'``:  like ``'periodic'``-padding but gives the smallest
            possible number of decomposition coefficients.
            Only available with ``impl='pywt'``, See `pywt.MODES.modes`.

        pad_const : float, optional
            Constant value to use if ``pad_mode == 'constant'``. Ignored
            otherwise. Constants other than 0 are not supported by the
            ``pywt`` back-end.
        impl : {'pywt'}, optional
            Back-end for the wavelet transform.

        Examples
        --------
        Check that the inverse is the actual inverse on a simple example on
        a discrete 2D space with 4 sampling points per axis:

        >>> space = odl.uniform_discr([0, 0], [1, 1], (4, 4))
        >>> wavelet_trafo = odl.trafos.WaveletTransform(
        ...     domain=space, nlevels=1, wavelet='haar')
        >>> orig_array = np.array([[1, 1, 1, 1],
        ...                        [0, 0, 0, 0],
        ...                        [0, 0, 1, 1],
        ...                        [1, 0, 1, 0]])
        >>> decomp = wavelet_trafo(orig_array)
        >>> recon = wavelet_trafo.inverse(decomp)
        >>> np.allclose(recon, orig_array)
        True
        """
        super().__init__(space=range, wavelet=wavelet, nlevels=nlevels,
                         variant='inverse', pad_mode=pad_mode,
                         pad_const=pad_const, impl=impl)

    def _call(self, coeffs):
        """Return the inverse wavelet transform of ``coeffs``."""
        if self.impl == 'pywt':
            shapes = pywt_coeff_shapes(self.range.shape, self.pywt_wavelet,
                                       self.nlevels, self.pywt_pad_mode)
            coeff_list = pywt_coeffs_from_flat_array(coeffs, shapes)
            return pywt_multi_level_recon(
                coeff_list, recon_shape=self.range.shape,
                wavelet=self.pywt_wavelet, mode=self.pywt_pad_mode)
        else:
            raise RuntimeError("bad `impl` '{}'".format(self.impl))

    @property
    def adjoint(self):
        """Adjoint of this operator.

        Returns
        -------
        adjoint : `WaveletTransform`
            If the transform is orthogonal, the adjoint is the inverse.

        Raises
        ------
        OpNotImplementedError
            if `is_orthogonal` is ``False``

        See Also
        --------
        inverse
        """
        if self.is_orthogonal:
            return self.inverse
        else:
            # TODO: put adjoint here
            return super().adjoint

    @property
    def inverse(self):
        """Inverse of this operator.

        Returns
        -------
        inverse : `WaveletTransform`

        See Also
        --------
        adjoint
        """
        return WaveletTransform(
            domain=self.range, nlevels=self.nlevels, wavelet=self.pywt_wavelet,
            pad_mode=self.pad_mode, pad_const=self.pad_const, impl=self.impl)


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests(skip_if=not PYWT_AVAILABLE)
